from typing import Dict
import numpy as np
from tdw.output_data import StaticRobot
from tdw.robot_data.drive import Drive
from tdw.robot_data.joint_type import JointType


class JointStatic:
    """
    Static robot joint data.
    """

    def __init__(self, static_robot: StaticRobot, joint_index: int):
        """
        :param static_robot: Static robot output data from the build.
        :param joint_index: The index of this joint in the list of joints.
        """

        """:field
        The ID of this joint.
        """
        self.joint_id: int = static_robot.get_joint_id(joint_index)
        """:field
        The name of the joint.
        """
        self.name: str = static_robot.get_joint_name(joint_index)
        """:field
        [The type of joint.](joint_type.md)
        """
        self.joint_type: JointType = JointType[static_robot.get_joint_type(joint_index)]
        """:field
        The segmentation color of this joint as an `[r, g, b]` numpy array.
        """
        self.segmentation_color: np.array = np.array(static_robot.get_joint_segmentation_color(joint_index))
        """:field
        The mass of this joint.
        """
        self.mass: float = static_robot.get_joint_mass(joint_index)
        """:field
        If True, this joint is immovable.
        """
        self.immovable: bool = static_robot.get_is_joint_immovable(joint_index)
        """:field
        If True, this is the root joint.
        """
        self.root: bool = static_robot.get_is_joint_root(joint_index)
        """:field
        The ID of this joint's parent joint. Ignore if `self.root == True`.
        """
        self.parent_id: int = static_robot.get_joint_parent_id(joint_index)
        """:field
        A dictionary of [Drive data](drive.md) for each of the robot's joints. Key = The drive axis (`"x"`, `"y"`, or `"z"`).
        """
        self.drives: Dict[str, Drive] = dict()
        for i in range(static_robot.get_num_joint_drives(joint_index)):
            drive = Drive(sr=static_robot, joint_index=joint_index, drive_index=i)
            self.drives[drive.axis] = drive



import zmq
import json
import os
from subprocess import Popen
from typing import List, Union, Tuple, Dict
from tdw.librarian import ModelLibrarian, SceneLibrarian, MaterialLibrarian, HDRISkyboxLibrarian, \
    HumanoidAnimationLibrarian, HumanoidLibrarian, HumanoidAnimationRecord, RobotLibrarian
from tdw.backend.paths import EDITOR_LOG_PATH, PLAYER_LOG_PATH
from tdw.output_data import Version, QuitSignal
from tdw.release.build import Build
from tdw.release.pypi import PyPi
from tdw.version import __version__
from tdw.add_ons.add_on import AddOn
from tdw.physics_audio.object_audio_static import DEFAULT_OBJECT_AUDIO_STATIC_DATA
from tdw.physics_audio.audio_material import AudioMaterial
from tdw.physics_audio.audio_material_constants import STATIC_FRICTION, DYNAMIC_FRICTION, DENSITIES


class Controller:
    """
    Base class for all controllers.

    Usage:

    ```python
    from tdw.controller import Controller
    c = Controller()
    ```
    """

    MODEL_LIBRARIANS: Dict[str, ModelLibrarian] = dict()
    SCENE_LIBRARIANS: Dict[str, SceneLibrarian] = dict()
    MATERIAL_LIBRARIANS: Dict[str, MaterialLibrarian] = dict()
    HDRI_SKYBOX_LIBRARIANS: Dict[str, HDRISkyboxLibrarian] = dict()
    HUMANOID_LIBRARIANS: Dict[str, HumanoidLibrarian] = dict()
    HUMANOID_ANIMATION_LIBRARIANS: Dict[str, HumanoidAnimationLibrarian] = dict()
    ROBOT_LIBRARIANS: Dict[str, RobotLibrarian] = dict()

    def __init__(self, port: int = 1071, check_version: bool = True, launch_build: bool = True):
        """
        Create the network socket and bind the socket to the port.

        :param port: The port number.
        :param check_version: If true, the controller will check the version of the build and print the result.
        :param launch_build: If True, automatically launch the build. If one doesn't exist, download and extract the correct version. Set this to False to use your own build, or (if you are a backend developer) to use Unity Editor.
        """

        # A list of modules that will add commands on `communicate()`.
        self.add_ons: List[AddOn] = list()

        # Compare the installed version of the tdw Python module to the latest on PyPi.
        # If there is a difference, recommend an upgrade.
        if check_version:
            self._check_pypi_version()

        # Launch the build.
        if launch_build:
            Controller.launch_build(port=port)
        context = zmq.Context()
        # noinspection PyUnresolvedReferences
        self.socket = context.socket(zmq.REP)
        self.socket.bind('tcp://*:' + str(port))

        self.socket.recv()

        # Set error handling to default values (the build will try to quit on errors and exceptions).
        # Request the version to log it and remember here if the Editor is being used.
        resp = self.communicate([{"$type": "set_error_handling"},
                                 {"$type": "send_version"},
                                 {"$type": "load_scene",
                                  "scene_name": "ProcGenScene"}])
        self._is_standalone: bool = False
        self._tdw_version: str = ""
        self._unity_version: str = ""
        for r in resp[:-1]:
            if Version.get_data_type_id(r) == "vers":
                v = Version(r)
                self._tdw_version = v.get_tdw_version()
                self._unity_version = v.get_unity_version()
                self._is_standalone = v.get_standalone()
                break
        # Compare the version of the tdw module to the build version.
        if check_version and launch_build:
            self._check_build_version()

    def communicate(self, commands: Union[dict, List[dict]]) -> list:
        """
        Send commands and receive output data in response.

        :param commands: A list of JSON commands.

        :return The output data from the build.
        """

        if isinstance(commands, dict):
            commands = [commands]

        # Append commands from each add-on.
        for m in self.add_ons:
            # Initialize an add-on.
            if not m.initialized:
                commands.extend(m.get_initialization_commands())
                m.initialized = True
            # Append the add-on's commands.
            else:
                commands.extend(m.commands)
                m.commands.clear()
        # Possibly do something with the commands about to be sent.
        for m in self.add_ons:
            m.before_send(commands)

        # Serialize the message.
        msg = [json.dumps(commands).encode('utf-8')]
        # Send the commands.
        self.socket.send_multipart(msg)
        # Receive output data.
        resp = self.socket.recv_multipart()

        # Occasionally, the build's socket will stop receiving messages.
        # If that happens, it will close the socket, create a new socket, and send a dummy output data object.
        # The ID of the dummy object is "ftre" (FailedToReceive).
        # If the controller receives the dummy object, it should re-send its commands.
        # The dummy object is always in an array: [ftre, 0]
        # This way, the controller can easily differentiate it from a response that just has the frame count.
        ftre: bool = True
        num_ftre: int = 0
        while ftre and num_ftre < 1000:
            ftre = False
            for i in range(len(resp) - 1):
                if resp[i][4:8] == b'ftre':
                    ftre = True
                    self.socket.send_multipart(msg)
                    resp = self.socket.recv_multipart()
                    num_ftre += 1
                    break
        # Tried too many times.
        if ftre:
            print("Quitting now because the controller tried too many times to resend commands to the build. "
                  "Check the build log for more info.")
            self._print_build_log()

        # Check if we've received a quit signal. If we have, check if there was an error.
        for i in range(len(resp) - 1):
            if resp[i][4:8] == b'quit':
                if not QuitSignal(resp[i]).get_ok():
                    print("The build quit due to an error. Check the build log for more info.")
                    self._print_build_log()
                break

        # Get commands per module for the next frame.
        for m in self.add_ons:
            m.on_send(resp=resp)

        # Return the output data from the build.
        return resp

    @staticmethod
    def get_add_object(model_name: str, object_id: int, position: Dict[str, float] = None, rotation: Dict[str, float] = None, library: str = "") -> dict:
        """
        Returns a valid add_object command.

        :param model_name: The name of the model.
        :param position: The position of the model. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The starting rotation of the model, in Euler angles. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param library: The path to the records file. If left empty, the default library will be selected. See `ModelLibrarian.get_library_filenames()` and `ModelLibrarian.get_default_library()`.
        :param object_id: The ID of the new object.

        :return An add_object command that the controller can then send.
        """

        if library == "":
            library = "models_core.json"
        if library not in Controller.MODEL_LIBRARIANS:
            Controller.MODEL_LIBRARIANS[library] = ModelLibrarian(library)
        record = Controller.MODEL_LIBRARIANS[library].get_record(model_name)

        return {"$type": "add_object",
                "name": model_name,
                "url": record.get_url(),
                "scale_factor": record.scale_factor,
                "position": position if position is not None else {"x": 0, "y": 0, "z": 0},
                "rotation": rotation if rotation is not None else {"x": 0, "y": 0, "z": 0},
                "category": record.wcategory,
                "id": object_id}

    @staticmethod
    def get_add_physics_object(model_name: str, object_id: int, position: Dict[str, float] = None, rotation: Dict[str, float] = None, library: str = "", scale_factor: Dict[str, float] = None, kinematic: bool = False, gravity: bool = True, default_physics_values: bool = True, mass: float = 1, dynamic_friction: float = 0.3, static_friction: float = 0.3, bounciness: float = 0.7) -> List[dict]:
        """
        Add an object to the scene with physics values (mass, friction coefficients, etc.).

        :param model_name: The name of the model.
        :param position: The position of the model. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The starting rotation of the model, in Euler angles. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param library: The path to the records file. If left empty, the default library will be selected. See `ModelLibrarian.get_library_filenames()` and `ModelLibrarian.get_default_library()`.
        :param object_id: The ID of the new object.
        :param scale_factor: The [scale factor](../api/command_api.md#scale_object).
        :param kinematic: If True, the object will be [kinematic](../api/command_api.md#set_kinematic_state).
        :param gravity: If True, the object won't respond to [gravity](../api/command_api.md#set_kinematic_state).
        :param default_physics_values: If True, use default physics values. Not all objects have default physics values. To determine if object does: `has_default_physics_values = model_name in DEFAULT_OBJECT_AUDIO_STATIC_DATA`.
        :param mass: The mass of the object. Ignored if `default_physics_values == True`.
        :param dynamic_friction: The [dynamic friction](../api/command_api.md#set_physic_material) of the object. Ignored if `default_physics_values == True`.
        :param static_friction: The [static friction](../api/command_api.md#set_physic_material) of the object. Ignored if `default_physics_values == True`.
        :param bounciness: The [bounciness](../api/command_api.md#set_physic_material) of the object. Ignored if `default_physics_values == True`.

        :return: A list of commands to add the object and apply physics values.
        """

        if library == "":
            library = "models_core.json"
        if library not in Controller.MODEL_LIBRARIANS:
            Controller.MODEL_LIBRARIANS[library] = ModelLibrarian(library)
        if position is None:
            position = {"x": 0, "y": 0, "z": 0}
        record = Controller.MODEL_LIBRARIANS[library].get_record(model_name)
        if position is None:
            position = {"x": 0, "y": 0, "z": 0}
        commands = [{"$type": "add_object",
                     "name": record.name,
                     "url": record.get_url(),
                     "scale_factor": record.scale_factor,
                     "position": position,
                     "category": record.wcategory,
                     "id": object_id}]
        if rotation is not None:
            # The rotation is a quaternion.
            if "w" in rotation:
                commands.append({"$type": "rotate_object_to",
                                 "rotation": rotation,
                                 "id": object_id})
            # The rotation is in Euler angles.
            else:
                commands.append({"$type": "rotate_object_to_euler_angles",
                                 "euler_angles": rotation,
                                 "id": object_id})
        if scale_factor is not None:
            commands.append({"$type": "scale_object",
                             "scale_factor": scale_factor,
                             "id": object_id})
        commands.append({"$type": "set_kinematic_state",
                         "id": object_id,
                         "is_kinematic": kinematic,
                         "use_gravity": gravity})
        # Kinematic objects must be continuous_speculative.
        if kinematic:
            commands.append({"$type": "set_object_collision_detection_mode",
                             "id": object_id,
                             "mode": "continuous_speculative"})

        if default_physics_values:
            # Use default physics values.
            if model_name in DEFAULT_OBJECT_AUDIO_STATIC_DATA:
                mass = DEFAULT_OBJECT_AUDIO_STATIC_DATA[model_name].mass
                bounciness = DEFAULT_OBJECT_AUDIO_STATIC_DATA[model_name].bounciness
                material = DEFAULT_OBJECT_AUDIO_STATIC_DATA[model_name].material
            # Fallback: Try to derive physics values from existing data.
            else:
                if "models_full.json" not in Controller.MODEL_LIBRARIANS:
                    Controller.MODEL_LIBRARIANS["models_full.json"] = ModelLibrarian("models_full.json")
                # Get all models in the same category that have default physics values.
                records = Controller.MODEL_LIBRARIANS["models_full.json"].get_all_models_in_wnid(record.wnid)
                records = [r for r in records if not r.do_not_use and r.name != record.name and r.name in
                           DEFAULT_OBJECT_AUDIO_STATIC_DATA]
                # Fallback: Find objects with similar volume.
                if len(records) == 0:
                    records = [r for r in Controller.MODEL_LIBRARIANS["models_full.json"].records if r.name in
                               DEFAULT_OBJECT_AUDIO_STATIC_DATA and not r.do_not_use and r.name != record.name and
                               0.8 <= abs(r.volume / record.volume) <= 1.2]
                # Fallback: Select a default material and bounciness.
                if len(records) == 0:
                    material: AudioMaterial = AudioMaterial.plastic_hard
                    # Select a default bounciness.
                    bounciness: float = 0
                # Select the most common material and bounciness.
                else:
                    materials: List[AudioMaterial] = [DEFAULT_OBJECT_AUDIO_STATIC_DATA[r.name].material for r in records]
                    material: AudioMaterial = max(set(materials), key=materials.count)
                    bouncinesses = [DEFAULT_OBJECT_AUDIO_STATIC_DATA[r.name].bounciness for r in records]
                    bounciness = round(sum(bouncinesses) / len(bouncinesses), 3)
                # Derive the mass.
                mass = DENSITIES[material] * record.volume
            commands.extend([{"$type": "set_mass",
                              "mass": mass,
                              "id": object_id},
                             {"$type": "set_physic_material",
                              "dynamic_friction": DYNAMIC_FRICTION[material],
                              "static_friction": STATIC_FRICTION[material],
                              "bounciness": bounciness,
                              "id": object_id}])
        else:
            commands.extend([{"$type": "set_mass",
                              "mass": mass,
                              "id": object_id},
                             {"$type": "set_physic_material",
                              "dynamic_friction": dynamic_friction,
                              "static_friction": static_friction,
                              "bounciness": bounciness,
                              "id": object_id}])
        return commands

    @staticmethod
    def get_add_material(material_name: str, library: str = "") -> dict:
        """
        Returns a valid add_material command.

        :param material_name: The name of the material.
        :param library: The path to the records file. If left empty, the default library will be selected. See `MaterialLibrarian.get_library_filenames()` and `MaterialLibrarian.get_default_library()`.

        :return An add_material command that the controller can then send.
        """

        if library == "":
            library = "materials_med.json"
        if library not in Controller.MATERIAL_LIBRARIANS:
            Controller.MATERIAL_LIBRARIANS[library] = MaterialLibrarian(library)
        record = Controller.MATERIAL_LIBRARIANS[library].get_record(material_name)
        return {"$type": "add_material",
                "name": material_name,
                "url": record.get_url()}

    @staticmethod
    def get_add_scene(scene_name: str, library: str = "") -> dict:
        """
        Returns a valid add_scene command.

        :param scene_name: The name of the scene.
        :param library: The path to the records file. If left empty, the default library will be selected. See `SceneLibrarian.get_library_filenames()` and `SceneLibrarian.get_default_library()`.

        :return An add_scene command that the controller can then send.
        """

        if library == "":
            library = "scenes.json"
        if library not in Controller.SCENE_LIBRARIANS:
            Controller.SCENE_LIBRARIANS[library] = SceneLibrarian(library)
        record = Controller.SCENE_LIBRARIANS[library].get_record(scene_name)
        return {"$type": "add_scene",
                "name": scene_name,
                "url": record.get_url()}

    @staticmethod
    def get_add_hdri_skybox(skybox_name: str, library: str = "") -> dict:
        """
        Returns a valid add_hdri_skybox command.

        :param skybox_name: The name of the skybox.
        :param library: The path to the records file. If left empty, the default library will be selected. See `HDRISkyboxLibrarian.get_library_filenames()` and `HDRISkyboxLibrarian.get_default_library()`.

        :return An add_hdri_skybox command that the controller can then send.
        """

        if library == "":
            library = "hdri_skyboxes.json"
        if library not in Controller.HDRI_SKYBOX_LIBRARIANS:
            Controller.HDRI_SKYBOX_LIBRARIANS[library] = HDRISkyboxLibrarian(library)
        record = Controller.HDRI_SKYBOX_LIBRARIANS[library].get_record(skybox_name)
        return {"$type": "add_hdri_skybox",
                "name": skybox_name,
                "url": record.get_url(),
                "exposure": record.exposure,
                "initial_skybox_rotation": record.initial_skybox_rotation,
                "sun_elevation": record.sun_elevation,
                "sun_initial_angle": record.sun_initial_angle,
                "sun_intensity": record.sun_intensity}

    @staticmethod
    def get_add_humanoid(humanoid_name: str, object_id: int, position: Dict[str, float] = None, rotation: Dict[str, float] = None, library: str = "") -> dict:
        """
        Returns a valid add_humanoid command.

        :param humanoid_name: The name of the humanoid.
        :param position: The position of the humanoid.
        :param rotation: The starting rotation of the humanoid, in Euler angles.
        :param library: The path to the records file. If left empty, the default library will be selected. See `HumanoidLibrarian.get_library_filenames()` and `HumanoidLibrarian.get_default_library()`.
        :param object_id: The ID of the new object.

        :return An add_humanoid command that the controller can then send.
        """

        if position is None:
            position = {"x": 0, "y": 0, "z": 0}
        if rotation is None:
            rotation = {"x": 0, "y": 0, "z": 0}

        if library == "":
            library = "humanoids.json"
        if library not in Controller.HUMANOID_LIBRARIANS:
            Controller.HUMANOID_LIBRARIANS[library] = HumanoidLibrarian(library)
        record = Controller.HUMANOID_LIBRARIANS[library].get_record(humanoid_name)
        return {"$type": "add_humanoid",
                "name": humanoid_name,
                "url": record.get_url(),
                "position": position,
                "rotation": rotation,
                "id": object_id}

    @staticmethod
    def get_add_humanoid_animation(humanoid_animation_name: str, library="") -> (dict, HumanoidAnimationRecord):
        """
        Returns a valid add_humanoid_animation command and the record (which you will need to play an animation).

        :param humanoid_animation_name: The name of the animation.
        :param library: The path to the records file. If left empty, the default library will be selected. See `HumanoidAnimationLibrarian.get_library_filenames()` and `HumanoidAnimationLibrarian.get_default_library()`.

        :return An add_humanoid_animation command that the controller can then send.
        """

        if library == "":
            library = "humanoid_animations.json"
        if library not in Controller.HUMANOID_ANIMATION_LIBRARIANS:
            Controller.HUMANOID_ANIMATION_LIBRARIANS[library] = HumanoidAnimationLibrarian(library)
        record = Controller.HUMANOID_ANIMATION_LIBRARIANS[library].get_record(humanoid_animation_name)
        return {"$type": "add_humanoid_animation",
                "name": humanoid_animation_name,
                "url": record.get_url()}, record

    @staticmethod
    def get_add_robot(name: str, robot_id: int, position: Dict[str, float] = None, rotation: Dict[str, float] = None, library: str = "") -> dict:
        """
        Returns a valid add_robot command.

        :param name: The name of the robot.
        :param robot_id: A unique ID for the robot.
        :param position: The initial position of the robot. If None, the position will be (0, 0, 0).
        :param rotation: The initial rotation of the robot in Euler angles.
        :param library: The path to the records file. If left empty, the default library will be selected. See `RobotLibrarian.get_library_filenames()` and `RobotLibrarian.get_default_library()`.

        :return An `add_robot` command that the controller can then send.
        """

        if library == "":
            library = "robots.json"
        if library not in Controller.ROBOT_LIBRARIANS:
            Controller.ROBOT_LIBRARIANS[library] = RobotLibrarian(library)
        record = Controller.ROBOT_LIBRARIANS[library].get_record(name)

        if position is None:
            position = {"x": 0, "y": 0, "z": 0}
        if rotation is None:
            rotation = {"x": 0, "y": 0, "z": 0}

        assert record is not None, f"Robot metadata record not found: {name}"
        return {"$type": "add_robot",
                "id": robot_id,
                "position": position,
                "rotation": rotation,
                "name": name,
                "url": record.get_url()}

    def get_version(self) -> Tuple[str, str]:
        """
        Send a send_version command to the build.

        :return The TDW version and the Unity Engine version.
        """

        resp = self.communicate({"$type": "send_version"})
        for r in resp[:-1]:
            if Version.get_data_type_id(r) == "vers":
                v = Version(r)
                return v.get_tdw_version(), v.get_unity_version()
        if len(resp) == 1:
            raise Exception("Tried receiving version output data but didn't receive anything!")
        raise Exception(f"Expected output data with ID version but got: " + Version.get_data_type_id(resp[0]))

    @staticmethod
    def get_unique_id() -> int:
        """
        Generate a unique integer. Useful when creating objects.

        :return The new unique ID.
        """

        return int.from_bytes(os.urandom(3), byteorder='big')

    @staticmethod
    def get_frame(frame: bytes) -> int:
        """
        Converts the frame byte array to an integer.

        :param frame: The frame as bytes.

        :return The frame as an integer.
        """

        return int.from_bytes(frame, byteorder='big')

    @staticmethod
    def launch_build(port: int = 1071) -> None:
        """
        Launch the build. If a build doesn't exist at the expected location, download one to that location.

        :param port: The socket port.
        """

        # Download the build.
        need_to_download = False
        if not Build.BUILD_PATH.exists():
            print(f"Couldn't find build at {Build.BUILD_PATH}\nDownloading now...")
            need_to_download = True
        else:
            # Check versions.
            build_version_path = Build.BUILD_ROOT_DIR.joinpath("TDW/version.txt")
            if build_version_path.exists():
                build_version = build_version_path.read_text().strip()
            else:
                build_version = "(unknown!)"
            if build_version != __version__:
                print(f"Python version is {__version__} but the build version is {build_version}.\n"
                      f"Downloading version {__version__} of the build now...")
                need_to_download = True

        # Download a new version of the build.
        if need_to_download:
            success = Build.download()
            if not success:
                print("You need to launch your own build.")
        else:
            success = True
        # Launch the build.
        if success:
            Popen([str(Build.BUILD_PATH.resolve()), "-port "+str(port)])

    def _check_build_version(self, version: str = __version__, build_version: str = None) -> None:
        """
        Check the version of the build. If there is no build, download it.
        If the build is of the wrong version, recommend an upgrade.

        :param version: The version of TDW. You can set this to an arbitrary version for testing purposes.
        :param build_version: If not None, this overrides the expected build version. Only override for debugging.
        """

        # Override the build version for testing.
        if build_version is not None:
            self._tdw_version = build_version
        print(f"Build version {self._tdw_version}\nUnity Engine {self._unity_version}\n"
              f"Python tdw module version {version}")

    def _print_build_log(self) -> None:
        """
        Print a message indicating where the build log is located.
        """

        if self._is_standalone:
            log_path = PLAYER_LOG_PATH
        else:
            log_path = EDITOR_LOG_PATH
        print(f"If the build is on the same machine as this controller, the log path is probably "
              f"{str(log_path.resolve())}")
        print(f"If the build is on a remote Linux server, the log path is probably"
              f" ~/.config/unity3d/MIT/TDW/Player.log (where ~ is your home directory)")

    @staticmethod
    def _check_pypi_version(v_installed_override: str = None, v_pypi_override: str = None) -> None:
        """
        Compare the version of the tdw Python module to the latest on PyPi.
        If there is a mismatch, offer an upgrade recommendation.

        :param v_installed_override: Override for the installed version. Change this to debug.
        :param v_pypi_override: Override for the PyPi version. Change this to debug.
        """

        # Get the version of the installed tdw module.
        installed_tdw_version = PyPi.get_installed_tdw_version()
        # Get the latest version of the tdw module on PyPi.
        pypi_version = PyPi.get_pypi_version()

        # Apply overrides
        if v_installed_override is not None:
            installed_tdw_version = v_installed_override
        if v_pypi_override is not None:
            pypi_version = v_pypi_override

        # If there is a mismatch, recommend an upgrade.
        if installed_tdw_version != pypi_version:
            # Strip the installed version of the post-release suffix (e.g. 1.6.3.4 to 1.6.3).
            stripped_installed_version = PyPi.strip_post_release(installed_tdw_version)
            # This message is here only for debugging.
            if stripped_installed_version != __version__:
                print(f"Your installed version: {stripped_installed_version} "
                      f"doesn't match tdw.version.__version__: {__version__} "
                      f"(this may be because you're using code from the tdw repo that is ahead of PyPi).")
            # Strip the latest PyPi version of the post-release suffix.
            stripped_pypi_version = PyPi.strip_post_release(pypi_version)
            print(f"You are using TDW {installed_tdw_version} but version {pypi_version} is available.")

            # If user is behind by a post release, recommend an upgrade to the latest.
            # (Example: installed version is 1.6.3.4 and PyPi version is 1.6.3.5)
            if stripped_installed_version == stripped_pypi_version:
                print(f"Upgrade to the latest version of TDW:\npip3 install tdw -U")

            # Using a version behind the latest (e.g. latest is 1.6.3 and installed is 1.6.2)
            # If the user is behind by a major or minor release, recommend either upgrading to a minor release
            # or to a major release.
            # (Example: installed version is 1.6.3.4 and PyPi version is 1.7.0.0)
            else:
                installed_major = PyPi.get_major_release(stripped_installed_version)
                pypi_minor = PyPi.get_latest_minor_release(stripped_installed_version)
                # Minor release mis-match.
                if PyPi.strip_post_release(pypi_minor) != stripped_installed_version:
                    print(f"To upgrade to the last version of 1.{installed_major}:\n"
                          f"pip3 install tdw=={pypi_minor}")
                pypi_major = PyPi.get_major_release(stripped_pypi_version)
                # Major release mis-match.
                if installed_major != pypi_major:
                    # Offer to upgrade to a major release.
                    print(f"Consider upgrading to the latest version of TDW ({stripped_pypi_version}):"
                          f"\npip3 install tdw -U")
        else:
            print("Your installed tdw Python module is up to date with PyPi.")



from enum import Enum
from typing import List, Dict, Optional, Union
import numpy as np
from tdw.tdw_utils import TDWUtils
from tdw.quaternion_utils import QuaternionUtils
from tdw.output_data import OutputData, AvatarKinematic, ImageSensors
from tdw.add_ons.third_person_camera_base import ThirdPersonCameraBase


class _MoveTargetType(Enum):
    """
    A type of target that the camera can move towards.
    """

    position = 1
    object = 2


class _RotateTargetType(Enum):
    """
    A type of target that the camera can rotate towards.
    """

    position = 1
    object = 2
    rotation = 4


class CinematicCamera(ThirdPersonCameraBase):
    """
    Wrapper class for third-person camera controls in TDW. These controls are "cinematic" in the sense that the camera will move, rotate, etc. *towards* a target at a set speed per frame. The `CinematicCamera` class is suitable for demo videos of TDW, but *not* for most actual experiments.
    """

    def __init__(self, avatar_id: str = None, position: Dict[str, float] = None, rotation: Dict[str, float] = None,
                 field_of_view: int = None, move_speed: float = 0.1, rotate_speed: float = 1,
                 field_of_view_speed: float = 0.1, look_at: Union[int, Dict[str, float]] = None):
        """
        :param avatar_id: The ID of the avatar (camera). If None, a random ID is generated.
        :param position: The initial position of the object.If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The initial rotation of the camera. Can be Euler angles (keys are `(x, y, z)`) or a quaternion (keys are `(x, y, z, w)`). If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param field_of_view: If not None, set the field of view.
        :param move_speed: The directional speed of the camera. This can later be adjusted by setting `self.move_speed`.
        :param rotate_speed: The angular speed of the camera. This can later be adjusted by setting `self.rotate_speed`.
        :param look_at: If not None, the cinematic camera will look at this object (if int) or position (if dictionary).
        :param field_of_view_speed: Adjust the field of view by this value per frame.
        """

        super().__init__(avatar_id=avatar_id, position=position, rotation=rotation, field_of_view=field_of_view)
        """:field
        The directional speed of the camera.
        """
        self.move_speed: float = move_speed
        """:field
        The angular speed of the camera.
        """
        self.rotate_speed: float = rotate_speed
        """:field
        Adjust the field of view by this value per frame.
        """
        self.field_of_view_speed: float = field_of_view_speed
        # The current forward directional vector of the image sensor.
        self._sensor_forward: np.array = np.array([0, 0, 0])
        # The current rotation of the image sensor.
        self._sensor_rotation: np.array = np.array([0, 0, 0, 0])

        # A target object ID or position to move towards. Can be None (no target).
        self._move_target: Optional[Union[int, Dict[str, float]]] = None
        # If `self._move_target` is an int, try to stay this far away from the target object.
        self._move_to_object_offset: Dict[str, float] = {"x": 0, "y": 0, "z": 0}
        # The type of move target.
        self._move_target_type: _MoveTargetType = _MoveTargetType.position

        # A target object ID, position, quaternion, or Euler angles to rotate to. Can be None (no target).
        self._rotate_target: Optional[Union[int, Dict[str, float]]] = None
        # The type of rotate target.
        self._rotate_target_type: _RotateTargetType = _RotateTargetType.position

        self._look_at: Optional[Union[int, Dict[str, float]]] = look_at

        # The target field of view.
        self._field_of_view_target: float = self._field_of_view

    def get_initialization_commands(self) -> List[dict]:
        commands = super().get_initialization_commands()
        commands.extend([{"$type": "send_avatars",
                          "frequency": "always"},
                         {"$type": "send_image_sensors",
                          "frequency": "always"}])
        if self._look_at is not None:
            # Look at the object.
            if isinstance(self._look_at, int):
                commands.append({"$type": "look_at",
                                 "object_id": self._look_at,
                                 "use_centroid": True,
                                 "avatar_id": self.avatar_id})
                # Continue to look at the object.
                self.rotate_to_object(target=self._look_at)
            # Look at the position.
            elif isinstance(self._look_at, dict):
                commands.append({"$type": "look_at_position",
                                 "position": self._look_at,
                                 "avatar_id": self.avatar_id})
                # Continue to look at the position.
                self.rotate_to_position(target=self._look_at)
            else:
                raise TypeError(f"Invalid look-at target: {self._look_at}")
        return commands

    def on_send(self, resp: List[bytes]) -> None:
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "imse":
                imse = ImageSensors(resp[i])
                if imse.get_avatar_id() == self.avatar_id:
                    # Update the rotation.
                    self._sensor_forward = np.array(imse.get_sensor_forward(0))
                    self._sensor_rotation = np.array(imse.get_sensor_rotation(0))
                    # Set the field of view.
                    self._field_of_view = imse.get_sensor_field_of_view(0)
                    set_field_of_view = False
                    if self._field_of_view_target is None:
                        self._field_of_view_target = self._field_of_view
                    f = np.linalg.norm(self._field_of_view - self._field_of_view_target)
                    if f > self.field_of_view_speed:
                        if self._field_of_view > self._field_of_view_target:
                            self._field_of_view -= self.field_of_view_speed
                        else:
                            self._field_of_view += self.field_of_view_speed
                        set_field_of_view = True
                    else:
                        if f < 1e-5 and f < self.field_of_view_speed:
                            self._field_of_view = self._field_of_view_target
                            set_field_of_view = True
                    if set_field_of_view:
                        self.commands.append({"$type": "set_field_of_view",
                                              "field_of_view": self._field_of_view,
                                              "avatar_id": self.avatar_id})
            elif r_id == "avki":
                a = AvatarKinematic(resp[i])
                if a.get_avatar_id() == self.avatar_id:
                    # Update the position.
                    self.position = TDWUtils.array_to_vector3(np.array(a.get_position()))
        if self._move_target is not None:
            if self._move_target_type == _MoveTargetType.position:
                self.commands.append({"$type": "move_avatar_towards_position",
                                      "position": self._move_target,
                                      "speed": self.move_speed,
                                      "avatar_id": self.avatar_id})
            elif self._move_target_type == _MoveTargetType.object:
                self.commands.append({"$type": "move_avatar_towards_object",
                                      "object_id": self._move_target,
                                      "offset": self._move_to_object_offset,
                                      "use_centroid": True,
                                      "speed": self.move_speed,
                                      "avatar_id": self.avatar_id})
            else:
                raise Exception(f"Invalid move target type: {self._move_target_type}")
        # Rotate towards a target.
        if self._rotate_target is not None:
            if self._rotate_target_type == _RotateTargetType.rotation:
                self.commands.append({"$type": "rotate_sensor_container_towards_rotation",
                                      "avatar_id": self.avatar_id,
                                      "rotation": self._rotate_target,
                                      "speed": self.rotate_speed})
            elif self._rotate_target_type == _RotateTargetType.object:
                self.commands.append({"$type": "rotate_sensor_container_towards_object",
                                      "avatar_id": self.avatar_id,
                                      "speed": self.rotate_speed,
                                      "object_id": self._rotate_target,
                                      "use_centroid": True})
            elif self._rotate_target_type == _RotateTargetType.position:
                self.commands.append({"$type": "rotate_sensor_container_towards_position",
                                      "avatar_id": self.avatar_id,
                                      "position": self._rotate_target,
                                      "speed": self.rotate_speed})
            else:
                raise Exception(f"Invalid rotate target type: {self._move_target_type}")

    def move_to_position(self, target: Dict[str, float], relative: bool = False) -> None:
        """
        Start moving towards a target position.

        :param relative: If True, the target is relative to the current position of the avatar. If False, the target is in absolute worldspace coordinates.
        :param target: The target position.
        """

        if relative:
            self._move_target = TDWUtils.array_to_vector3(TDWUtils.vector3_to_array(target) +
                                                          TDWUtils.vector3_to_array(self.position))
        else:
            self._move_target = target
        self._move_target_type = _MoveTargetType.position

    def move_to_object(self, target: int, offset: Dict[str, float]) -> None:
        """
        Start moving towards a target object.

        :param target: The ID of the target object.
        :param offset: Stop moving when the camera is this far away from the object.
        """

        self._move_target = target
        self._move_to_object_offset = offset
        self._move_target_type = _MoveTargetType.object

    def stop_moving(self) -> None:
        """
        Stop moving towards the current target.
        """

        self._move_target = None

    def rotate_to_object(self, target: int) -> None:
        """
        Rotate towards an object. This will update if

        :param target: The ID of the target object.
        """

        self._rotate_target = target
        self._rotate_target_type = _RotateTargetType.object

    def rotate_to_position(self, target: Dict[str, float]) -> None:
        """
        Start to rotate towards a position.

        :param target: The target position.
        """

        self._rotate_target = target
        self._rotate_target_type = _RotateTargetType.position

    def rotate_by_rpy(self, target: Dict[str, float]) -> None:
        """
        Rotate the camera by the `[pitch, yaw, roll]` angles expressed as an `[x, y, z]` dictionary.

        :param target: The target `[pitch, yaw, roll]` angles from when this function was first called, in degrees.
        """

        eulers = QuaternionUtils.quaternion_to_euler_angles(self._sensor_rotation)
        eulers[0] += target["x"]
        eulers[1] += target["y"]
        eulers[2] += target["z"]
        self._rotate_target = TDWUtils.array_to_vector4(
            QuaternionUtils.euler_angles_to_quaternion(np.deg2rad(eulers)))
        self._rotate_target_type = _RotateTargetType.rotation

    def rotate_to_rotation(self, target: Dict[str, float]) -> None:
        """
        Rotate towards a rotation quaternion.

        :param target: The target rotation.
        """

        self._rotate_target = target
        self._rotate_target_type = _RotateTargetType.rotation

    def stop_rotating(self) -> None:
        """
        Stop rotating towards the current target.
        """

        self._rotate_target = None

    def set_field_of_view(self, field_of_view: float) -> None:
        """
        Set the target field of view. This will also set the camera's target focal length.

        :param field_of_view: The field of view.
        """

        self._field_of_view_target = field_of_view



from tdw.add_ons.audio_initializer_base import AudioInitializerBase


class AudioInitializer(AudioInitializerBase):
    """
    Initialize standard (Unity) audio.

    This assumes that an avatar corresponding to `avatar_id` has already been added to the scene.
    """

    def __init__(self, avatar_id: str = "a", framerate: int = 60):
        """
        :param avatar_id: The ID of the listening avatar.
        :param framerate: The target simulation framerate.
        """

        super().__init__(avatar_id=avatar_id, framerate=framerate)

    def _get_sensor_command_name(self) -> str:
        return "add_audio_sensor"

    def _get_play_audio_command_name(self) -> str:
        return "play_audio_data"



from typing import List
from tdw.add_ons.add_on import AddOn


class StepPhysics(AddOn):
    """
    Step n + 1 physics frames per communciate() call.
    """

    def __init__(self, num_frames: int):
        """
        :param num_frames: Step this many physics frames + 1 per communicate() call.
        """

        super().__init__()
        """:field
        Step this many physics frames + 1 per communicate() call.
        """
        self.num_frames: int = num_frames

    def get_initialization_commands(self) -> List[dict]:
        return []

    def on_send(self, resp: List[bytes]) -> None:
        self.commands.append({"$type": "step_physics",
                              "frames": self.num_frames})



import numpy as np


class ObjectStatic:
    """
    Static data for an object. This data won't change between frames.
    """

    def __init__(self, name: str, object_id: int, mass: float, segmentation_color: np.array, size: np.array,
                 category: str, kinematic: bool, dynamic_friction: float, static_friction: float, bounciness: float):
        """
        :param name: The name of the object.
        :param object_id: The unique ID of the object.
        :param mass: The mass of the object.
        :param segmentation_color: The segmentation color of the object.
        :param size: The size of the object.
        :param dynamic_friction: The dynamic friction of the object.
        :param static_friction: The static friction of the object.
        :param bounciness: The bounciness of the object.
        :param kinematic: If True, this object is kinematic, and won't respond to physics.
        """

        """:field
        The unique ID of the object.
        """
        self.object_id: int = object_id
        """:field
        [The name of the model.](https://github.com/threedworld-mit/tdw/blob/master/Documentation/python/librarian/model_librarian.md)
        """
        self.name: str = name.lower()
        """:field
         The semantic category of the object.
         """
        self.category: str = category
        """:field
        If True, this object is kinematic, and won't respond to physics. 
        """
        self.kinematic = kinematic
        """:field
        The RGB segmentation color for the object as a numpy array: `[r, g, b]`
        """
        self.segmentation_color = segmentation_color
        """:field
        The mass of the object.
        """
        self.mass = mass
        """:field
        The size of the object as a numpy array: `[width, height, length]`
        """
        self.size = size
        """:field
        The dynamic friction of the object.
        """
        self.dynamic_friction: float = dynamic_friction
        """:field
        The static friction of the object.
        """
        self.static_friction: float = static_friction
        """:field
        The bounciness of the object.
        """
        self.bounciness: float = bounciness



import numpy as np
from tdw.output_data import Robot


class JointDynamic:
    """
    Dynamic info for a joint that can change per-frame, such as its current position.
    """

    def __init__(self, robot: Robot, joint_index: int):
        """
        :param robot: The `Robot` output data from the build.
        :param joint_index: The index of the data for this joint.
        """

        """:field
        The ID of this joint.
        """
        self.joint_id: int = robot.get_joint_id(joint_index)
        """:field
        The worldspace position of this joint as an `[x, y, z]` numpy array.
        """
        self.position: np.array = np.array(robot.get_joint_position(joint_index))
        """:field
        The angles of each axis of the joint in degrees. For prismatic joints, you need to convert this from degrees to radians in order to get the correct distance in meters.
        """
        self.angles: np.array = robot.get_joint_positions(joint_index)
        """:field
        If True, this joint is currently moving.
        """
        self.moving: bool = False



from typing import List
import time
from threading import Thread
import socket as sock
from contextlib import closing
import zmq


class RemoteBuildLauncher:
    """
    Connect to a remote binary_manager daemon and launch an instance of a TDW build.
    """

    @staticmethod
    def launch_build(listener_port: int, build_address: str, controller_address: str) -> dict:
        """
        Connect to a remote binary_manager daemon and launch an instance of a TDW build.

        Returns the necessary information for a local controller to connect.
        Use this function to automatically launching binaries on remote (or local) nodes, and to
        automatically shut down the build after controller is finished. Call in the constructor
        of a controller and pass the build_port returned in build_info to the parent Controller class.

        :param listener_port: The port launch_binaries is listening on.
        :param build_address: Remote IP or hostname of node running launch_binaries.
        :param controller_address: IP or hostname of node running controller.

        :return The build_info dictionary containing build_port.
        """

        context = zmq.Context()
        # noinspection PyUnresolvedReferences
        socket = context.socket(zmq.REQ)
        socket.connect("tcp://" + build_address + ":%s" % listener_port)
        build_info = RemoteBuildLauncher._send_start_build(socket, controller_address)
        thread = Thread(target=RemoteBuildLauncher._keep_alive_thread,
                        args=(socket, build_info))
        thread.setDaemon(True)
        thread.start()
        return build_info

    @staticmethod
    def get_unity_args(arg_dict: dict) -> List[str]:
        """
        :param arg_dict: A dictionary of arguments. Key=The argument prefix (e.g. port) Value=Argument value.

        :return The formatted command line string that is accepted by unity arg parser.
        """

        formatted_args = []
        for key, value in arg_dict.items():
            prefix = "-" + key + "="
            if type(value) == list:
                prefix += ",".join([str(v) for v in value])
            else:
                prefix += str(value)
            formatted_args += [prefix]
        return formatted_args

    @staticmethod
    def find_free_port() -> int:
        """
        :return A free socket port.
        """

        with closing(sock.socket(sock.AF_INET, sock.SOCK_STREAM)) as s:
            s.bind(("", 0))
            return int(s.getsockname()[1])

    @staticmethod
    def _send_start_build(socket, controller_address: str) -> dict:
        """
        This sends a command to the launch_binaries daemon running on a remote node
        to start a binary connected to the given controller address.

        :param socket: The zmq socket.
        :param controller_address: The host name or ip address of node running the controller.

        :return Build info dictionary containing build port.
        """
        request = {"type": "start_build",
                   "controller_address": controller_address}
        socket.send_json(request)
        build_info = socket.recv_json()
        return build_info

    @staticmethod
    def _send_keep_alive(socket, build_info: dict) -> dict:
        """
        This sends a command to the launch_binaries daemon running on a remote node
        to mark a given binary as still alive, preventing garbage collection.

        :param socket: The zmq socket.
        :param build_info: A diciontary containing the build_port.

        :return a heartbeat indicating build is still alive.
        """

        build_port = build_info["build_port"]
        request = {"type": "keep_alive", "build_port": build_port}
        socket.send_json(request)
        heartbeat = socket.recv_json()
        return heartbeat

    @staticmethod
    def _send_kill_build(socket, build_info: dict) -> dict:
        """
        This sends a command to the launch_binaries daemon running on a remote node to terminate a given binary.

        :param socket: The zmq socket.
        :param build_info: A diciontary containing the build_port.

        :return A kill_status indicating build has been succesfully terminated.
        """

        build_port = build_info["build_port"]
        request = {"type": "kill_build", "build_port": build_port}
        socket.send_json(request)
        kill_status = socket.recv_json()
        return kill_status

    @staticmethod
    def _keep_alive_thread(socket, build_info: dict) -> None:
        """
        This is a wrapper around the keep alive command to be executed in a separate thread.

        :param socket: The zmq socket.
        :param build_info: A diciontary containing the build_port.
        """
        while True:
            RemoteBuildLauncher._send_keep_alive(socket, build_info)
            time.sleep(60)



import numpy as np


class Rigidbody:
    """
    Dynamic object rigidbody data. Note that this excludes *static* rigidbody data such as the mass of the object.
    """

    def __init__(self, velocity: np.array, angular_velocity: np.array, sleeping: bool):
        """
        :param velocity: The directional velocity of the object.
        :param angular_velocity: The angular velocity of the object.
        :param sleeping: If True, the object isn't moving.
        """

        """:field
        The directional velocity of the object.
        """
        self.velocity: np.array = velocity
        """:field
        The angular velocity of the object.
        """
        self.angular_velocity: np.array = angular_velocity
        """:field
        If True, the object isn't moving.
        """
        self.sleeping: bool = sleeping



from os import urandom
import base64
import math
import json
from pathlib import Path
from pkg_resources import resource_filename
from typing import Dict, Optional, Union, List, Tuple
import numpy as np
import scipy.signal as sg
from scipy.ndimage import gaussian_filter1d
from pydub import AudioSegment
from tdw.output_data import OutputData, Rigidbodies, StaticRobot, SegmentationColors, StaticRigidbodies, \
    RobotJointVelocities
from tdw.physics_audio.audio_material import AudioMaterial
from tdw.physics_audio.object_audio_static import ObjectAudioStatic, DEFAULT_OBJECT_AUDIO_STATIC_DATA
from tdw.physics_audio.modes import Modes
from tdw.physics_audio.base64_sound import Base64Sound
from tdw.physics_audio.collision_audio_info import CollisionAudioInfo
from tdw.physics_audio.collision_audio_type import CollisionAudioType
from tdw.physics_audio.collision_audio_event import CollisionAudioEvent
from tdw.physics_audio.scrape_model import ScrapeModel, DEFAULT_SCRAPE_MODELS
from tdw.physics_audio.scrape_material import ScrapeMaterial
from tdw.object_data.rigidbody import Rigidbody
from tdw.audio_constants import SAMPLE_RATE, CHANNELS, SAMPLE_WIDTH
from tdw.add_ons.collision_manager import CollisionManager
from tdw.librarian import MaterialLibrarian


class PyImpact(CollisionManager):
    """
    Generate impact sounds from physics data. Sounds can be synthesized automatically (for general use-cases) or manually (for advanced use-cases).

    ```python
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.add_ons.audio_initializer import AudioInitializer
    from tdw.add_ons.py_impact import PyImpact

    c = Controller()
    commands = [TDWUtils.create_empty_room(12, 12)]
    commands.extend(TDWUtils.create_avatar(avatar_id="a",
                                           position={"x": 1, "y": 1.6, "z": -2},
                                           look_at={"x": 0, "y": 0.5, "z": 0}))
    commands.extend(c.get_add_physics_object(model_name="vase_02",
                                             position={"x": 0, "y": 3, "z": 0},
                                             object_id=c.get_unique_id()))
    audio_initializer = AudioInitializer(avatar_id="a")
    py_impact = PyImpact()
    c.add_ons.extend([audio_initializer, py_impact])
    c.communicate(commands)
    for i in range(200):
        c.communicate([])
    c.communicate({"$type": "terminate"})
    ```

    When using PyImpact, please cite  [Traer,Cusimano and McDermott, A perceptually inspired generative model of rigid-body contact sounds, Digital Audio Effects, (DAFx), 2019](http://dafx2019.bcu.ac.uk/papers/DAFx2019_paper_57.pdf) and [Agarwal, Cusimano, Traer, and McDermott, Object-based synthesis of scraping and rolling sounds based on non-linear physical constraints, (DAFx), 2021](http://mcdermottlab.mit.edu/bib2php/pubs/makeAbs.php?loc=agarwal21).

    ```
    @article {4500,
        title = {A perceptually inspired generative model of rigid-body contact sounds},
        journal = {Proceedings of the 22nd International Conference on Digital Audio Effects (DAFx-19)},
        year = {2019},
        month = {09/2019},
        abstract = {<p>Contact between rigid-body objects produces a diversity of impact and friction sounds. These sounds can be synthesized with detailed simulations of the motion, vibration and sound radiation of the objects, but such synthesis is computationally expensive and prohibitively slow for many applications. Moreover, detailed physical simulations may not be necessary for perceptually compelling synthesis; humans infer ecologically relevant causes of sound, such as material categories, but not with arbitrary precision. We present a generative model of impact sounds which summarizes the effect of physical variables on acoustic features via statistical distributions fit to empirical measurements of object acoustics. Perceptual experiments show that sampling from these distributions allows efficient synthesis of realistic impact and scraping sounds that convey material, mass, and motion.</p>
    },
        author = {James Traer and Maddie Cusimano and Josh H. McDermott}
    }
    ```

    ```
    @inproceedings{agarwal21,
         TITLE= "Object-based synthesis of scraping and rolling sounds based on non-linear physical constraints",
         AUTHOR= "V Agarwal and M Cusimano and J Traer and J H McDermott",
         booktitle= "The 24th International Conference on Digital Audio Effects (DAFx-21)",
         MONTH= "September",
         YEAR= 2021,
         PDF-URL= "http://mcdermottlab.mit.edu/papers/Agarwal_etal_2021_scraping_rolling_synthesis_DAFx.pdf",
    }
    ```
    """

    """:class_var
    50ms of silence. Used for scrapes.
    """
    SILENCE_50MS: AudioSegment = AudioSegment.silent(duration=50, frame_rate=SAMPLE_RATE)
    """:class_var
    The maximum velocity allowed for a scrape.
    """
    SCRAPE_MAX_VELOCITY: float = 5.0
    """:class_var
    Meters per pixel on the scrape surface.
    """
    SCRAPE_M_PER_PIXEL: float = 1394.068 * 10 ** -9
    """:class_var
    The default amp value for objects.
    """
    DEFAULT_AMP: float = 0.2
    """:class_var
    The default [material](../physics_audio/audio_material.md) for objects.
    """
    DEFAULT_MATERIAL: AudioMaterial = AudioMaterial.plastic_hard
    """:class_var
    The default resonance value for objects.
    """
    DEFAULT_RESONANCE: float = 0.45
    """:class_var
    The default audio size "bucket" for objects.
    """
    DEFAULT_SIZE: int = 1
    """:class_var
    The assumed bounciness value for robot joints.
    """
    ROBOT_JOINT_BOUNCINESS: float = 0.6
    """:class_var
    The [material](../physics_audio/audio_material.md) used for robot joints.
    """
    ROBOT_JOINT_MATERIAL: AudioMaterial = AudioMaterial.metal
    """:class_var
    The amp value for the floor.
    """
    FLOOR_AMP: float = 0.5
    """:class_var
    The size "bucket" for the floor.
    """
    FLOOR_SIZE: int = 4
    """:class_var
    The mass of the floor.
    """
    FLOOR_MASS: int = 100
    # Visual material librarian used for scrape surfaces.
    __VISUAL_MATERIAL_LIBRARIAN: MaterialLibrarian = MaterialLibrarian("materials_high.json")

    def __init__(self, initial_amp: float = 0.5, prevent_distortion: bool = True, logging: bool = False,
                 static_audio_data_overrides: Dict[int, ObjectAudioStatic] = None,
                 resonance_audio: bool = False, floor: AudioMaterial = AudioMaterial.wood_medium,
                 rng: np.random.RandomState = None, auto: bool = True, scrape: bool = True,
                 scrape_objects: Dict[int, ScrapeModel] = None):
        """
        :param initial_amp: The initial amplitude, i.e. the "master volume". Must be > 0 and < 1.
        :param prevent_distortion: If True, clamp amp values to <= 0.99
        :param logging: If True, log mode properties for all colliding objects, as json.
        :param static_audio_data_overrides: If not None, a dictionary of audio data. Key = Object ID; Value = [`ObjectAudioStatic`](../physics_audio/object_audio_static.md). These audio values will be applied to these objects instead of default values.
        :param resonance_audio: If True, the simulation is using Resonance Audio.
        :param floor: The floor material.
        :param rng: The random number generator. If None, a random number generator with a random seed is created.
        :param auto: If True, PyImpact will evalulate the simulation state per `communicate()` call and automatically generate audio.
        :param scrape: If True, initialize certain objects as scrape surfaces: Change their visual material(s) and enable them for scrape audio. See: `tdw.physics_audio.scrape_model.DEFAULT_SCRAPE_MODELS`
        :param scrape_objects: If `scrape == True` and this is not None, this dictionary can be used to manually set scrape surfaces. Key = Object ID. Value = [`ScrapeModel`](../physics_audio/scrape_model.md).
        """

        super().__init__()

        if rng is None:
            """:field
            The random number generator.
            """
            self.rng: np.random.RandomState = np.random.RandomState()
        else:
            self.rng = rng

        assert 0 < initial_amp < 1, f"initial_amp is {initial_amp} (must be > 0 and < 1)."

        """:field
        The initial amplitude, i.e. the "master volume". Must be > 0 and < 1.
        """
        self.initial_amp = initial_amp
        """:field
        If True, clamp amp values to <= 0.99
        """
        self.prevent_distortion = prevent_distortion
        """:field
        If True, log mode properties for all colliding objects, as json.
        """
        self.logging = logging

        """:field
        The collision info per set of objects.
        """
        self.object_modes: Dict[int, Dict[int, CollisionAudioInfo]] = dict()
        """:field
        If True, the simulation is using Resonance Audio.
        """
        self.resonance_audio: bool = resonance_audio
        """:field
        The floor material.
        """
        self.floor: AudioMaterial = floor

        """:field
        Cached material data.
        """
        self.material_data: Dict[str, dict] = {}
        material_list = ["ceramic", "wood_hard", "wood_medium", "wood_soft", "metal", "glass", "paper", "cardboard",
                         "leather", "fabric", "plastic_hard", "plastic_soft_foam", "rubber", "stone"]
        for mat in material_list:
            for i in range(6):
                # Load the JSON data.
                mat_name = mat + "_" + str(i)
                path = mat_name + "_mm"
                data = json.loads(Path(resource_filename(__name__, f"py_impact/material_data/{path}.json")).read_text())
                self.material_data.update({mat_name: data})
        """:field
        Cached scrape surface data.
        """
        self.scrape_surface_data: Dict[ScrapeMaterial, Dict[str, np.ndarray]] = {}
        """:field
        A dictionary of all [scrape models](../physics_audio/scrape_model.md) in the scene. If `scrape == False`, this dictionary is empty. Key = Object ID.
        """
        self._scrape_objects: Dict[int, ScrapeModel] = dict()
        self._scrape: bool = scrape
        # Use scrape surfaces.
        if self._scrape and scrape_objects is not None:
            for k in scrape_objects:
                self._scrape_objects[k] = scrape_objects[k]
        """:field
        The mode properties log.
        """
        self.mode_properties_log = dict()
        """:field
        If True, PyImpact will evalulate the simulation state per `communicate()` call and automatically generate audio.
        """
        self.auto: bool = auto
        # A dictionary of audio data. Key = Object ID; Value = `ObjectAudioStatic`.
        # These audio values will be applied to these objects instead of default values.
        self._static_audio_data_overrides: Dict[int, ObjectAudioStatic] = dict()
        if static_audio_data_overrides is not None:
            self._static_audio_data_overrides = static_audio_data_overrides
        """:field
        Collision events on this frame. Key = Object ID. Value = [`CollisionAudioEvent`](../physics_audio/collision_audio_event.md).
        """
        self.collision_events: Dict[int, CollisionAudioEvent] = dict()

        self._cached_audio_info: bool = False
        # A dictionary of audio data. Key = Object ID; Value = `ObjectAudioStatic`.
        self._static_audio_data: Dict[int, ObjectAudioStatic] = dict()

        # Summed scrape masters. Key = primary ID, secondary ID.
        self._scrape_summed_masters: Dict[Tuple[int, int], AudioSegment] = dict()
        # Keeping a track of previous scrape indices.
        self._scrape_previous_indices: Dict[Tuple[int, int], int] = dict()
        # Starting velocity magnitude of scraping object; use in calculating changing band-pass filter.
        self._scrape_start_velocities: Dict[Tuple[int, int], float] = dict()
        # Initialize the scraping event counter.
        self._scrape_events_count: Dict[Tuple[int, int], int] = dict()

    def get_initialization_commands(self) -> List[dict]:
        return [{"$type": "send_rigidbodies",
                 "frequency": "always"},
                {"$type": "send_robot_joint_velocities",
                 "frequency": "always"},
                {"$type": "send_collisions",
                 "enter": True,
                 "exit": True,
                 "stay": True,
                 "collision_types": ["obj", "env"]},
                {"$type": "send_static_robots"},
                {"$type": "send_segmentation_colors"},
                {"$type": "send_static_rigidbodies"}]

    def on_send(self, resp: List[bytes]) -> None:
        super().on_send(resp=resp)
        # Cache static audio info.
        if not self._cached_audio_info:
            self._cached_audio_info = True
            self._cache_static_data(resp=resp)
        # Don't automatically generate audio.
        if not self.auto:
            return
        # Get collision events.
        self._get_collision_types(resp=resp)
        for object_id in self.collision_events:
            command = None
            # Generate an impact sound.
            if self.collision_events[object_id].collision_type == CollisionAudioType.impact:
                # Generate an environment sound.
                if self.collision_events[object_id].secondary_id is None:
                    audio = self._static_audio_data[object_id]
                    command = self.get_impact_sound_command(velocity=self.collision_events[object_id].velocity,
                                                            contact_points=self.collision_events[object_id].collision.points,
                                                            contact_normals=self.collision_events[object_id].collision.normals,
                                                            primary_id=object_id,
                                                            primary_amp=audio.amp,
                                                            primary_material=audio.material.name + "_" + str(audio.size),
                                                            primary_mass=audio.mass,
                                                            secondary_id=None,
                                                            secondary_amp=PyImpact.FLOOR_AMP,
                                                            secondary_material=self._get_floor_material_name(),
                                                            secondary_mass=PyImpact.FLOOR_MASS,
                                                            primary_resonance=audio.resonance,
                                                            secondary_resonance=audio.resonance)
                # Generate an object sound.
                else:
                    target_audio = self._static_audio_data[self.collision_events[object_id].primary_id]
                    other_audio = self._static_audio_data[self.collision_events[object_id].secondary_id]
                    command = self.get_impact_sound_command(velocity=self.collision_events[object_id].velocity,
                                                            contact_points=self.collision_events[object_id].collision.points,
                                                            contact_normals=self.collision_events[object_id].collision.normals,
                                                            primary_id=target_audio.object_id,
                                                            primary_amp=target_audio.amp,
                                                            primary_material=target_audio.material.name + "_" + str(
                                                                target_audio.size),
                                                            primary_mass=target_audio.mass,
                                                            secondary_id=other_audio.object_id,
                                                            secondary_amp=other_audio.amp,
                                                            secondary_material=other_audio.material.name + "_" + str(
                                                                other_audio.size),
                                                            secondary_mass=other_audio.mass,
                                                            primary_resonance=target_audio.resonance,
                                                            secondary_resonance=other_audio.resonance)
            # Generate a scrape sound.
            elif self.collision_events[object_id].collision_type == CollisionAudioType.scrape and self.collision_events[object_id].secondary_id in self._scrape_objects:
                scrape_surface_id = self.collision_events[object_id].secondary_id
                # Generate an object sound.
                if scrape_surface_id is not None:
                    target_audio = self._static_audio_data[self.collision_events[object_id].primary_id]
                    other_audio = self._static_audio_data[self.collision_events[object_id].secondary_id]
                    command = self.get_scrape_sound_command(velocity=self.collision_events[object_id].velocity,
                                                            contact_points=self.collision_events[object_id].collision.points,
                                                            contact_normals=self.collision_events[object_id].collision.normals,
                                                            primary_id=target_audio.object_id,
                                                            primary_amp=target_audio.amp,
                                                            primary_material=self._scrape_objects[scrape_surface_id].audio_material.name + "_" + str(target_audio.size),
                                                            primary_mass=target_audio.mass,
                                                            secondary_id=other_audio.object_id,
                                                            secondary_amp=other_audio.amp,
                                                            secondary_material=other_audio.material.name + "_" + str(other_audio.size),
                                                            secondary_mass=other_audio.mass,
                                                            primary_resonance=target_audio.resonance,
                                                            secondary_resonance=other_audio.resonance,
                                                            scrape_material=self._scrape_objects[scrape_surface_id].scrape_material)
            # Append impact sound commands.
            if command is not None:
                self.commands.append(command)

    def _get_floor_material_name(self) -> str:
        """
        :return: The name of the floor material.
        """

        # We probably need dedicated wall and floor materials, or maybe they are in size category #6?
        # Setting to "4" for now, for general debugging purposes.
        return f"{self.floor.name}_{PyImpact.FLOOR_SIZE}"

    def _get_collision_types(self, resp: List[bytes]) -> None:
        """
        Get all collision types on this frame. Update previous area data.

        :param resp: The response from the build.
        """

        # Collision events per object on this frame. We'll only care about the most significant one.
        collision_events_per_object: Dict[int, List[CollisionAudioEvent]] = dict()
        # Get the previous areas.
        previous_areas: Dict[int, float] = {k: v.area for k, v in self.collision_events.items()}
        # Clear the collision events.
        self.collision_events.clear()
        rigidbody_data: Dict[int, Rigidbody] = dict()
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            # Get rigidbody data.
            if r_id == "rigi":
                rigidbodies = Rigidbodies(resp[i])
                for j in range(rigidbodies.get_num()):
                    rigidbody_data[rigidbodies.get_id(j)] = Rigidbody(velocity=np.array(rigidbodies.get_velocity(j)),
                                                                      angular_velocity=np.array(
                                                                          rigidbodies.get_angular_velocity(j)),
                                                                      sleeping=rigidbodies.get_sleeping(j))
            # Get robot joint velocity data.
            elif r_id == "rojv":
                robot_joint_velocities = RobotJointVelocities(resp[i])
                for j in range(robot_joint_velocities.get_num_joints()):
                    rigidbody_data[robot_joint_velocities.get_joint_id(j)] = Rigidbody(velocity=robot_joint_velocities.get_joint_velocity(j),
                                                                                       angular_velocity=robot_joint_velocities.get_joint_angular_velocity(j),
                                                                                       sleeping=robot_joint_velocities.get_joint_sleeping(j))
        # Get collision data.
        for object_ids in self.obj_collisions:
            collider_id = object_ids.int1
            collidee_id = object_ids.int2
            event = CollisionAudioEvent(collision=self.obj_collisions[object_ids],
                                        object_0_static=self._static_audio_data[collider_id],
                                        object_0_dynamic=rigidbody_data[collider_id],
                                        object_1_static=self._static_audio_data[collidee_id],
                                        object_1_dynamic=rigidbody_data[collidee_id],
                                        previous_areas=previous_areas)
            # End an ongoing scrape, if any.
            if event.collision_type != CollisionAudioType.scrape:
                self._end_scrape((event.primary_id, event.secondary_id))
            if event.primary_id not in collision_events_per_object:
                collision_events_per_object[event.primary_id] = list()
            collision_events_per_object[event.primary_id].append(event)
        for object_id in self.env_collisions:
            event = CollisionAudioEvent(collision=self.env_collisions[object_id],
                                        object_0_static=self._static_audio_data[object_id],
                                        object_0_dynamic=rigidbody_data[object_id],
                                        previous_areas=previous_areas)
            # End an ongoing scrape, if any.
            if event.collision_type != CollisionAudioType.scrape:
                self._end_scrape((event.primary_id, event.secondary_id))
            if event.primary_id not in collision_events_per_object:
                collision_events_per_object[event.primary_id] = list()
            collision_events_per_object[event.primary_id].append(event)
        # Get the significant collision events per object.
        for primary_id in collision_events_per_object:
            events: List[CollisionAudioEvent] = [e for e in collision_events_per_object[primary_id] if e.magnitude > 0
                                                 and e.collision_type != CollisionAudioType.none]
            if len(events) > 0:
                event: CollisionAudioEvent = max(events, key=lambda x: x.magnitude)
                self.collision_events[event.primary_id] = event

    def _get_object_modes(self, material: Union[str, AudioMaterial]) -> Modes:
        """
        :param material: The audio material.

        :return: The audio modes.
        """
        data = self.material_data[material] if isinstance(material, str) else self.material_data[material.name]
        # Load the mode properties.
        f = -1
        p = -1
        t = -1
        for jm in range(0, 10):
            jf = 0
            while jf < 20:
                jf = data["cf"][jm] + self.rng.normal(0, data["cf"][jm] / 10)
            jp = data["op"][jm] + self.rng.normal(0, 10)
            jt = 0
            while jt < 0.001:
                jt = data["rt"][jm] + self.rng.normal(0, data["rt"][jm] / 10)
            if jm == 0:
                f = jf
                p = jp
                t = jt * 1e3
            else:
                f = np.append(f, jf)
                p = np.append(p, jp)
                t = np.append(t, jt * 1e3)
        return Modes(f, p, t)

    def get_impact_sound(self, velocity: np.array, contact_normals: List[np.array],
                         primary_id: int, primary_material: str, primary_amp: float, primary_mass: float,
                         secondary_id: Optional[int], secondary_material: str, secondary_amp: float,
                         secondary_mass: float, primary_resonance: float, secondary_resonance: float) -> Optional[Base64Sound]:
        """
        Produce sound of two colliding objects as a byte array.

        :param primary_id: The object ID for the primary (target) object.
        :param primary_material: The material label for the primary (target) object.
        :param secondary_id: The object ID for the secondary (other) object.
        :param secondary_material: The material label for the secondary (other) object.
        :param primary_amp: Sound amplitude of primary (target) object.
        :param secondary_amp: Sound amplitude of the secondary (other) object.
        :param primary_resonance: The resonance of the primary (target) object.
        :param secondary_resonance: The resonance of the secondary (other) object.
        :param velocity: The velocity.
        :param contact_normals: The collision contact normals.
        :param primary_mass: The mass of the primary (target) object.
        :param secondary_mass: The mass of the secondary (target) object.

        :return Sound data as a Base64Sound object.
        """

        # The sound amplitude of object 2 relative to that of object 1.
        amp2re1 = secondary_amp / primary_amp

        # Set the object modes.
        if secondary_id not in self.object_modes:
            self.object_modes.update({secondary_id: {}})
        if primary_id not in self.object_modes[secondary_id]:
            self.object_modes[secondary_id].update({primary_id: CollisionAudioInfo(self._get_object_modes(secondary_material),
                                                                                   self._get_object_modes(
                                                                                       primary_material),
                                                                                   amp=primary_amp * self.initial_amp)})
        # Unpack useful parameters.
        speed = np.square(velocity)
        speed = np.sum(speed)
        speed = math.sqrt(speed)
        nvel = velocity / np.linalg.norm(velocity)
        nspd = []
        for jc in range(len(contact_normals)):
            tmp = np.asarray(contact_normals[jc])
            tmp = tmp / np.linalg.norm(tmp)
            tmp = np.arccos(np.clip(np.dot(tmp, nvel), -1.0, 1.0))
            # Scale the speed by the angle (i.e. we want speed Normal to the surface).
            tmp = speed * np.cos(tmp)
            nspd.append(tmp)
        normal_speed = np.mean(nspd)
        mass = np.min([primary_mass, secondary_mass])

        # Re-scale the amplitude.
        if self.object_modes[secondary_id][primary_id].count == 0:
            # Sample the modes.
            sound, modes_1, modes_2 = self._make_impact_audio(amp2re1, mass,
                                                              mat1=primary_material,
                                                              mat2=secondary_material,
                                                              id1=primary_id,
                                                              id2=secondary_id,
                                                              primary_resonance=primary_resonance,
                                                              secondary_resonance=secondary_resonance)
            # Save collision info - we will need for later collisions.
            amp = self.object_modes[secondary_id][primary_id].amp
            self.object_modes[secondary_id][primary_id].init_speed = normal_speed
            self.object_modes[secondary_id][primary_id].obj1_modes = modes_1
            self.object_modes[secondary_id][primary_id].obj2_modes = modes_2

        else:
            amp = self.object_modes[secondary_id][primary_id].amp * normal_speed / self.object_modes[secondary_id][primary_id].init_speed
            # Adjust modes here so that two successive impacts are not identical.
            modes_1 = self.object_modes[secondary_id][primary_id].obj1_modes
            modes_2 = self.object_modes[secondary_id][primary_id].obj2_modes
            modes_1.powers = modes_1.powers + self.rng.normal(0, 2, len(modes_1.powers))
            modes_2.powers = modes_2.powers + self.rng.normal(0, 2, len(modes_2.powers))
            sound = PyImpact._synth_impact_modes(modes_1, modes_2, mass, primary_resonance, secondary_resonance)
            self.object_modes[secondary_id][primary_id].obj1_modes = modes_1
            self.object_modes[secondary_id][primary_id].obj2_modes = modes_2

        if self.logging:
            mode_props = dict()
            self._log_modes(self.object_modes[secondary_id][primary_id].count, mode_props, primary_id, secondary_id,
                            modes_1, modes_2, amp, primary_material, secondary_material)

        # On rare occasions, it is possible for PyImpact to fail to generate a sound.
        if sound is None:
            return None

        # Count the collisions.
        self.object_modes[secondary_id][primary_id].count_collisions()

        # Prevent distortion by clamping the amp.
        if self.prevent_distortion and np.abs(amp) > 0.99:
            amp = 0.99

        sound = amp * sound / np.max(np.abs(sound))
        return Base64Sound(sound)

    def get_impact_sound_command(self, velocity: np.array, contact_points: List[np.array],
                                 contact_normals: List[np.array], primary_id: int,
                                 primary_material: str, primary_amp: float, primary_mass: float,
                                 secondary_id: Optional[int], secondary_material: str, secondary_amp: float,
                                 secondary_mass: float, primary_resonance: float, secondary_resonance: float) -> Optional[dict]:
        """
        Create an impact sound, and return a valid command to play audio data in TDW.
        "target" should usually be the smaller object, which will play the sound.
        "other" should be the larger (stationary) object.

        :param primary_id: The object ID for the primary (target) object.
        :param primary_material: The material label for the primary (target) object.
        :param secondary_id: The object ID for the secondary (other) object.
        :param secondary_material: The material label for the secondary (other) object.
        :param primary_amp: Sound amplitude of primary (target) object.
        :param secondary_amp: Sound amplitude of the secondary (other) object.
        :param primary_resonance: The resonance of the primary (target) object.
        :param secondary_resonance: The resonance of the secondary (other) object.
        :param velocity: The velocity.
        :param contact_points: The collision contact points.
        :param contact_normals: The collision contact normals.
        :param primary_mass: The mass of the primary (target) object.
        :param secondary_mass: The mass of the secondary (target) object.

        :return A `play_audio_data` or `play_point_source_data` command that can be sent to the build via `Controller.communicate()`.
        """

        sound = self.get_impact_sound(velocity=velocity, contact_normals=contact_normals, primary_id=primary_id,
                                      primary_material=primary_material, primary_amp=primary_amp,
                                      primary_mass=primary_mass, secondary_id=secondary_id,
                                      secondary_material=secondary_material, secondary_amp=secondary_amp,
                                      secondary_mass=secondary_mass, primary_resonance=primary_resonance, secondary_resonance=secondary_resonance)
        if sound is not None:
            # Use the primary object ID for the audio source ID to prevent a droning effect.
            return self._get_audio_command(audio_source_id=primary_id, contact_points=contact_points, sound=sound)
        # If PyImpact failed to generate a sound (which is rare!), fail silently here.
        else:
            return None

    def _make_impact_audio(self, amp2re1: float, mass: float, id1: int, id2: int, primary_resonance: float,
                           secondary_resonance: float, mat1: str = 'cardboard', mat2: str = 'cardboard') -> (np.array, Modes, Modes):
        """
        Generate an impact sound.

        :param mat1: The material label for one of the colliding objects.
        :param mat2: The material label for the other object.
        :param amp2re1: The sound amplitude of object 2 relative to that of object 1.
        :param mass: The mass of the smaller of the two colliding objects.
        :param id1: The ID for the one of the colliding objects.
        :param id2: The ID for the other object.
        :param primary_resonance: The resonance of one of the colliding objects.
        :param secondary_resonance: The resonance of the other object.

        :return The sound, and the object modes.
        """

        # Unpack material names.
        for jmat in AudioMaterial:
            if mat1 == jmat:
                tmp1 = jmat
                mat1 = tmp1.name
            if mat2 == jmat:
                tmp2 = jmat
                mat2 = tmp2.name
        # Sample modes of object1.
        modes_1 = self.object_modes[id2][id1].obj1_modes
        modes_2 = self.object_modes[id2][id1].obj2_modes
        # Scale the two sounds as specified.
        modes_2.decay_times = modes_2.decay_times + 20 * np.log10(amp2re1)
        snth = PyImpact._synth_impact_modes(modes_1, modes_2, mass, primary_resonance, secondary_resonance)
        return snth, modes_1, modes_2

    def _get_impulse_response(self, velocity: np.array, contact_normals: List[np.array], primary_id: int,
                              primary_material: str, primary_amp: float, primary_mass: float,
                              secondary_id: int, secondary_material: str, secondary_amp: float, secondary_mass: float,
                              primary_resonance: float, secondary_resonance: float) -> Tuple[np.array, float]:
        """
        Generate an impulse response from the modes for two specified objects.

        :param primary_id: The object ID for the primary (target) object.
        :param primary_material: The material label for the primary (target) object.
        :param secondary_id: The object ID for the secondary (other) object.
        :param secondary_material: The material label for the secondary (other) object.
        :param primary_amp: Sound amplitude of primary (target) object.
        :param secondary_amp: Sound amplitude of the secondary (other) object.
        :param primary_resonance: The resonance of the primary (target) object.
        :param secondary_resonance: The resonances of the secondary (other) object.
        :param velocity: The velocity.
        :param contact_normals: The collision contact normals.
        :param primary_mass: The mass of the primary (target) object.
        :param secondary_mass: The mass of the secondary (target) object.

        :return The impulse response and the frequency.
        """

        self.get_impact_sound(velocity=velocity, contact_normals=contact_normals, primary_id=primary_id,
                              primary_material=primary_material, primary_amp=primary_amp, primary_mass=primary_mass,
                              secondary_id=secondary_id, secondary_material=secondary_material,
                              secondary_amp=secondary_amp, secondary_mass=secondary_mass, primary_resonance=primary_resonance, secondary_resonance=secondary_resonance)

        modes_1 = self.object_modes[secondary_id][primary_id].obj1_modes
        modes_2 = self.object_modes[secondary_id][primary_id].obj2_modes
        h1 = modes_1.sum_modes(resonance=primary_resonance)
        h2 = modes_2.sum_modes(resonance=secondary_resonance)
        h = Modes.mode_add(h1, h2)
        return h, min(modes_1.frequencies)

    def get_scrape_sound_command(self, velocity: np.array, contact_points: np.array,
                                 contact_normals: List[np.array], primary_id: int,
                                 primary_material: str, primary_amp: float, primary_mass: float,
                                 secondary_id: Optional[int], secondary_material: str, secondary_amp: float,
                                 secondary_mass: float, primary_resonance: float, secondary_resonance: float,
                                 scrape_material: ScrapeMaterial) -> Optional[dict]:
        """
        :param primary_id: The object ID for the primary (target) object.
        :param primary_material: The material label for the primary (target) object.
        :param secondary_id: The object ID for the secondary (other) object.
        :param secondary_material: The material label for the secondary (other) object.
        :param primary_amp: Sound amplitude of primary (target) object.
        :param secondary_amp: Sound amplitude of the secondary (other) object.
        :param primary_resonance: The resonance of the primary (target) object.
        :param secondary_resonance: The resonance of the secondary (other) object.
        :param velocity: The velocity.
        :param contact_points: The collision contact points.
        :param contact_normals: The collision contact normals.
        :param primary_mass: The mass of the primary (target) object.
        :param secondary_mass: The mass of the secondary (target) object.
        :param scrape_material: The [scrape material](../physics_audio/scrape_material.md).

        :return A command to play a scrape sound.
        """

        sound = self.get_scrape_sound(velocity=velocity,
                                      contact_normals=contact_normals,
                                      primary_id=primary_id,
                                      primary_material=primary_material,
                                      primary_amp=primary_amp,
                                      primary_mass=primary_mass,
                                      secondary_id=secondary_id,
                                      secondary_material=secondary_material,
                                      secondary_amp=secondary_amp,
                                      secondary_mass=secondary_mass,
                                      primary_resonance=primary_resonance,
                                      secondary_resonance=secondary_resonance,
                                      scrape_material=scrape_material)
        if sound is None:
            return None
        else:
            # Use random audio source IDs so that multiple scrape sound chunks can play at the same time.
            return self._get_audio_command(audio_source_id=int.from_bytes(urandom(3), byteorder='big'),
                                           contact_points=contact_points,
                                           sound=sound)

    def get_scrape_sound(self, velocity: np.array, contact_normals: List[np.array], primary_id: int,
                         primary_material: str, primary_amp: float, primary_mass: float,
                         secondary_id: int, secondary_material: str, secondary_amp: float, secondary_mass: float,
                         primary_resonance: float, secondary_resonance: float, scrape_material: ScrapeMaterial) -> Optional[Base64Sound]:
        """
        Create a scrape sound, and return a valid command to play audio data in TDW.
        "target" should usually be the smaller object, which will play the sound.
        "other" should be the larger (stationary) object.

        :param primary_id: The object ID for the primary (target) object.
        :param primary_material: The material label for the primary (target) object.
        :param secondary_id: The object ID for the secondary (other) object.
        :param secondary_material: The material label for the secondary (other) object.
        :param primary_amp: Sound amplitude of primary (target) object.
        :param secondary_amp: Sound amplitude of the secondary (other) object.
        :param primary_resonance: The resonance of the primary (target) object.
        :param secondary_resonance: The resonance of the secondary (other) object.
        :param velocity: The velocity.
        :param contact_normals: The collision contact normals.
        :param primary_mass: The mass of the primary (target) object.
        :param secondary_mass: The mass of the secondary (target) object.
        :param scrape_material: The [scrape material](../physics_audio/scrape_material.md).

        :return A [`Base64Sound`](../physics_audio/base64_sound.md) object or None if no sound.
        """

        scrape_key: Tuple[int, int] = (primary_id, secondary_id)
        if scrape_key not in self._scrape_previous_indices:
            self._scrape_previous_indices[scrape_key] = 0

        # Initialize scrape variables; if this is an in=process scrape, these will be replaced bu te stored values.
        summed_master = AudioSegment.silent(duration=0, frame_rate=SAMPLE_RATE)
        scrape_event_count = 0

        # Is this a new scrape?
        if scrape_key in self._scrape_summed_masters:
            summed_master = self._scrape_summed_masters[scrape_key]
            scrape_event_count = self._scrape_events_count[scrape_key]
        else:
            # No -- add initialized values to dictionaries.
            self._scrape_summed_masters[scrape_key] = summed_master
            self._scrape_events_count[scrape_key] = scrape_event_count

        # Get magnitude of velocity of the scraping object.
        mag = min(np.linalg.norm(velocity), PyImpact.SCRAPE_MAX_VELOCITY)

        # Cache the starting velocity.
        if scrape_event_count == 0:
            self._scrape_start_velocities[scrape_key] = mag

        # Map magnitude to gain level -- decrease in velocity = rise in negative dB, i.e. decrease in gain.
        db2 = 40 * np.log10(mag / PyImpact.SCRAPE_MAX_VELOCITY) - 1
        db1 = 20 * np.log10(mag / PyImpact.SCRAPE_MAX_VELOCITY) - 10

        # Get impulse response of the colliding objects. Amp values would normally come from objects.csv.
        # We also get the lowest-frequency IR mode, which we use to set the high-pass filter cutoff below.
        scraping_ir, min_mode_freq = self._get_impulse_response(velocity=velocity,
                                                                contact_normals=contact_normals,
                                                                primary_id=primary_id,
                                                                primary_material=primary_material,
                                                                primary_amp=primary_amp,
                                                                primary_mass=primary_mass,
                                                                secondary_id=secondary_id,
                                                                secondary_material=secondary_material,
                                                                secondary_amp=secondary_amp,
                                                                secondary_mass=secondary_mass,
                                                                primary_resonance=primary_resonance,
                                                                secondary_resonance=secondary_resonance)
        # Cache the scrape material.
        # Don't do this when PyImpact is initialized because scrape surfaces are large files!
        # We don't want them in memory all the time and they can be a bit slow to load.
        if scrape_material not in self.scrape_surface_data:
            scrape_surface = np.load(str(
                Path(resource_filename(__name__, f"py_impact/scrape_surfaces/{scrape_material.name}.npy")).resolve()))
            for i in range(2):
                scrape_surface = np.append(scrape_surface, scrape_surface)
            #   Load the surface texture as a 1D vector
            #   Create surface texture of desired length
            #   Calculate first and second derivatives by first principles
            #   Apply non-linearity on the second derivative
            #   Apply a variable Gaussian average
            #   Calculate the horizontal and vertical forces
            #   Convolve the force with the impulse response
            dsdx = (scrape_surface[1:] - scrape_surface[0:-1]) / PyImpact.SCRAPE_M_PER_PIXEL
            d2sdx2 = (dsdx[1:] - dsdx[0:-1]) / PyImpact.SCRAPE_M_PER_PIXEL
            rough_ratio = (np.std(scrape_surface) / (9 * 10 ** -4)) ** 0.85
            r_gain = 20 * np.log10(rough_ratio)
            self.scrape_surface_data[scrape_material] = {"dsdx": dsdx,
                                                         "d2sdx2": d2sdx2,
                                                         "surface": scrape_surface,
                                                         "r_gain": r_gain}
        dist = mag / 10
        num_pts = int(np.floor(dist / PyImpact.SCRAPE_M_PER_PIXEL) + 1)
        # No scrape.
        if num_pts < 1:
            self._end_scrape(scrape_key)
            return None
        # interpolate the surface slopes and curvatures based on the velocity magnitude
        final_ind = self._scrape_previous_indices[scrape_key] + num_pts

        vect1 = np.linspace(0, 1, num_pts)
        vect2 = np.linspace(0, 1, 4410)

        if final_ind > len(self.scrape_surface_data[scrape_material]["surface"]) - 1:
            self._scrape_previous_indices[scrape_key] = 0
            final_ind = num_pts
            slope_int = np.interp(vect2, vect1, self.scrape_surface_data[scrape_material]["dsdx"][
                                                self._scrape_previous_indices[scrape_key]:final_ind])
            curve_int = np.interp(vect2, vect1, self.scrape_surface_data[scrape_material]["d2sdx2"][
                                                self._scrape_previous_indices[scrape_key]:final_ind])
        else:
            slope_int = np.interp(vect2, vect1, self.scrape_surface_data[scrape_material]["dsdx"][
                                                self._scrape_previous_indices[scrape_key]:final_ind])
            curve_int = np.interp(vect2, vect1, self.scrape_surface_data[scrape_material]["d2sdx2"][
                                                self._scrape_previous_indices[scrape_key]:final_ind])

        self._scrape_previous_indices[scrape_key] = final_ind

        curve_int_tan = np.tanh(curve_int / (100 * primary_mass))

        d2_section = gaussian_filter1d(curve_int_tan, 6)

        vert_force = d2_section
        hor_force = slope_int

        t_force2 = vert_force / max(np.abs(vert_force))
        t_force1 = hor_force[:len(vert_force)]

        noise_seg1 = AudioSegment(t_force1.tobytes(),
                                  frame_rate=SAMPLE_RATE,
                                  sample_width=SAMPLE_WIDTH,
                                  channels=CHANNELS)
        noise_seg2 = AudioSegment(t_force2.tobytes(),
                                  frame_rate=SAMPLE_RATE,
                                  sample_width=SAMPLE_WIDTH,
                                  channels=CHANNELS)
        # Fade head and tail.
        noise_seg_fade1 = noise_seg1.fade_in(4).fade_out(4)
        noise_seg_fade2 = noise_seg2.fade_in(4).fade_out(4)
        # Convolve the band-pass filtered sound with the impulse response.
        conv1 = sg.fftconvolve(scraping_ir, noise_seg_fade1.get_array_of_samples())
        conv2 = sg.fftconvolve(scraping_ir, noise_seg_fade2.get_array_of_samples())

        # Again, we need this as an AudioSegment for overlaying with the previous frame's segment.
        # Convert to 16-bit integers for Unity, normalizing to make sure to minimize loss of precision from truncating floating values.
        normalized_noise_ints_conv1 = PyImpact._normalize_16bit_int(conv1)
        noise_seg_conv1 = AudioSegment(normalized_noise_ints_conv1.tobytes(),
                                       frame_rate=SAMPLE_RATE,
                                       sample_width=SAMPLE_WIDTH,
                                       channels=CHANNELS)

        normalized_noise_ints_conv2 = PyImpact._normalize_16bit_int(conv2)
        noise_seg_conv2 = AudioSegment(normalized_noise_ints_conv2.tobytes(),
                                       frame_rate=SAMPLE_RATE,
                                       sample_width=SAMPLE_WIDTH,
                                       channels=CHANNELS)

        # Gain-adjust the convolved segment using db value computed earlier.
        noise_seg_conv1 = noise_seg_conv1.apply_gain(db1)
        noise_seg_conv2 = noise_seg_conv2.apply_gain(db2)

        noise_seg_conv = noise_seg_conv1.overlay(noise_seg_conv2)
        # Apply roughness gain.
        noise_seg_conv = noise_seg_conv.apply_gain(self.scrape_surface_data[scrape_material]["r_gain"] + 4)

        # Pad the end of master with 50 ms of silence, the start of the current segment with (n * 50ms) of silence, and overlay.
        padded_current = (PyImpact.SILENCE_50MS * scrape_event_count) + noise_seg_conv
        summed_master = summed_master + PyImpact.SILENCE_50MS
        summed_master = summed_master.overlay(padded_current)

        # Extract 100ms "chunk" of sound to send over to Unity.
        start_idx = 100 * scrape_event_count
        unity_chunk = summed_master[-(len(summed_master) - start_idx):][:100]
        # Update stored summed waveform.
        self._scrape_summed_masters[scrape_key] = summed_master

        # Update scrape event count.
        scrape_event_count += 1
        self._scrape_events_count[scrape_key] = scrape_event_count
        # Scrape data is handled differently than impact data, so we'll create a dummy object first.
        sound = Base64Sound(np.array([0]))
        # Set the audio data.
        sound.wav_str = base64.b64encode(unity_chunk.raw_data).decode()
        sound.length = len(unity_chunk.raw_data)
        sound.bytes = unity_chunk.raw_data
        return sound

    @staticmethod
    def _synth_impact_modes(modes1: Modes, modes2: Modes, mass: float, primary_resonance: float, secondary_resonance: float) -> np.array:
        """
        Generate an impact sound from specified modes for two objects, and the mass of the smaller object.

        :param modes1: Modes of object 1. A numpy array with: column1=mode frequencies (Hz); column2=mode onset powers in dB; column3=mode RT60s in milliseconds;
        :param modes2: Modes of object 2. Formatted as modes1/modes2.
        :param mass: the mass of the smaller of the two colliding objects.
        :param primary_resonance: The resonance of the object 1.
        :param secondary_resonance: The resonance of the object 2.

        :return The impact sound.
        """

        h1 = modes1.sum_modes(resonance=primary_resonance)
        h2 = modes2.sum_modes(resonance=secondary_resonance)
        h = Modes.mode_add(h1, h2)
        if len(h) == 0:
            return None
        # Convolve with force, with contact time scaled by the object mass.
        max_t = 0.001 * mass
        # A contact time over 2ms is unphysically long.
        max_t = np.min([max_t, 2e-3])
        n_pts = int(np.ceil(max_t * 44100))
        tt = np.linspace(0, np.pi, n_pts)
        frc = np.sin(tt)
        x = sg.fftconvolve(h, frc)
        x = x / abs(np.max(x))
        return x

    def reset(self, initial_amp: float = 0.5, static_audio_data_overrides: Dict[int, ObjectAudioStatic] = None,
              scrape_objects: Dict[int, ScrapeModel] = None) -> None:
        """
        Reset PyImpact. This is somewhat faster than creating a new PyImpact object per trial.

        :param initial_amp: The initial amplitude, i.e. the "master volume". Must be > 0 and < 1.
        :param static_audio_data_overrides: If not None, a dictionary of audio data. Key = Object ID; Value = [`ObjectAudioStatic`](../physics_audio/object_audio_static.md). These audio values will be applied to these objects instead of default values.
        :param scrape_objects: A dictionary of [scrape objects](../physics_audio/scrape_model.md) in the scene. Key = Object ID. Ignored if None or `scrape == False` in the constructor.
        """

        assert 0 < initial_amp < 1, f"initial_amp is {initial_amp} (must be > 0 and < 1)."
        self._cached_audio_info = False
        self.initialized = False
        self._static_audio_data.clear()
        self._static_audio_data_overrides.clear()
        # Use scrape surfaces.
        if self._scrape and scrape_objects is not None:
            for k in scrape_objects:
                self._scrape_objects[k] = scrape_objects[k]
        if static_audio_data_overrides is not None:
            for k in static_audio_data_overrides:
                self._static_audio_data_overrides[k] = static_audio_data_overrides[k]
        # Clear the object data.
        self.object_modes.clear()
        # Clear collision data.
        self.collision_events.clear()
        # Clear scrape data.
        self._scrape_summed_masters.clear()
        self._scrape_start_velocities.clear()
        self._scrape_events_count.clear()
        self._scrape_objects.clear()
        self._scrape_previous_indices.clear()
        # Stop all ongoing audio.
        self.commands.append({"$type": "stop_all_audio"})

    def _log_modes(self, count: int, mode_props: dict, id1: int, id2: int, modes_1: Modes, modes_2: Modes, amp: float, mat1: str, mat2: str):
        """
        Log mode properties info for a single collision event.

        :param count: Mode count for this material-material collision.
        :param mode_props: Dictionary to log to.
        :param id1: ID of the "other" object.
        :param id2: ID of the "target" object.
        :param modes_1: Modes of the "other" object.
        :param modes_2: Modes of the "target" object.
        :param amp: Adjusted amplitude value of collision.
        :param mat1: Material of the "other" object.
        :param mat2: Material of the "target" object.
        """

        mode_props["modes_count"] = count
        mode_props["other_id"] = id1
        mode_props["target_id"] = id2
        mode_props["amp"] = amp
        mode_props["other_material"] = mat1
        mode_props["target_material"] = mat2
        mode_props["modes_1.frequencies"] = modes_1.frequencies.tolist()
        mode_props["modes_1.powers"] = modes_1.powers.tolist()
        mode_props["modes_1.decay_times"] = modes_1.decay_times.tolist()
        mode_props["modes_2.frequencies"] = modes_2.frequencies.tolist()
        mode_props["modes_2.powers"] = modes_2.powers.tolist()
        mode_props["modes_2.decay_times"] = modes_2.decay_times.tolist()
        self.mode_properties_log[str(id1) + "_" + str(id2) + "__" + str(count)] = mode_props

    def _cache_static_data(self, resp: List[bytes]) -> None:
        """
        Cache static data.

        :param resp: The response from the build.
        """

        # Load the default object info.
        categories: Dict[int, str] = dict()
        names: Dict[int, str] = dict()
        robot_joints: Dict[int, dict] = dict()
        object_masses: Dict[int, float] = dict()
        object_bouncinesses: Dict[int, float] = dict()
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "segm":
                segm = SegmentationColors(resp[i])
                for j in range(segm.get_num()):
                    object_id = segm.get_object_id(j)
                    model_name = segm.get_object_name(j).lower()
                    names[object_id] = model_name
                    categories[object_id] = segm.get_object_category(j)
                    # Enable a scrape surface.
                    if self._scrape and (model_name in DEFAULT_SCRAPE_MODELS or object_id in self._scrape_objects):
                        self._scrape_objects[object_id] = DEFAULT_SCRAPE_MODELS[model_name]
                        # Add the visual material.
                        material_record = PyImpact.__VISUAL_MATERIAL_LIBRARIAN.get_record(
                            name=self._scrape_objects[object_id].visual_material)
                        self.commands.append({"$type": "add_material",
                                              "name": material_record.name,
                                              "url": material_record.get_url()})
                        # Set the visual material.
                        for sub_object in self._scrape_objects[object_id].sub_objects:
                            self.commands.append({"$type": "set_visual_material",
                                                  "material_index": sub_object.material_index,
                                                  "material_name": material_record.name,
                                                  "object_name": sub_object.name,
                                                  "id": object_id})
            elif r_id == "srob":
                srob = StaticRobot(resp[i])
                for j in range(srob.get_num_joints()):
                    joint_id = srob.get_joint_id(j)
                    robot_joints[joint_id] = {"name": srob.get_joint_name(j),
                                              "mass": srob.get_joint_mass(j)}
            elif r_id == "srig":
                srig = StaticRigidbodies(resp[i])
                for j in range(srig.get_num()):
                    object_masses[srig.get_id(j)] = srig.get_mass(j)
                    object_bouncinesses[srig.get_id(j)] = srig.get_bounciness(j)
        need_to_derive: List[int] = list()
        for object_id in names:
            name = names[object_id]
            # Use override data.
            if object_id in self._static_audio_data_overrides:
                self._static_audio_data[object_id] = self._static_audio_data_overrides[object_id]
                self._static_audio_data[object_id].mass = object_masses[object_id]
                self._static_audio_data[object_id].object_id = object_id
            # Use default audio data.
            elif name in DEFAULT_OBJECT_AUDIO_STATIC_DATA:
                self._static_audio_data[object_id] = DEFAULT_OBJECT_AUDIO_STATIC_DATA[name]
                self._static_audio_data[object_id].mass = object_masses[object_id]
                self._static_audio_data[object_id].object_id = object_id
            else:
                need_to_derive.append(object_id)
        current_values = self._static_audio_data.values()
        derived_data: Dict[int, ObjectAudioStatic] = dict()
        for object_id in need_to_derive:
            # Fallback option: comparable objects in the same category.
            objects_in_same_category = [o for o in categories if categories[o] == categories[object_id]]
            if len(objects_in_same_category) > 0:
                amps: List[float] = [a.amp for a in current_values]
                materials: List[AudioMaterial] = [a.material for a in current_values]
                resonances: List[float] = [a.resonance for a in current_values]
                sizes: List[int] = [a.size for a in current_values]
            # Fallback option: Find objects with similar volume.
            else:
                amps: List[float] = list()
                materials: List[AudioMaterial] = list()
                resonances: List[float] = list()
                sizes: List[int] = list()
                for m_id in object_masses:
                    if m_id == object_id or m_id not in self._static_audio_data:
                        continue
                    if np.abs(object_masses[m_id] / object_masses[object_id]) < 1.5:
                        amps.append(self._static_audio_data[m_id].amp)
                        materials.append(self._static_audio_data[m_id].material)
                        resonances.append(self._static_audio_data[m_id].resonance)
                        sizes.append(self._static_audio_data[m_id].size)
            # Fallback option: Use default values.
            if len(amps) == 0:
                amp: float = PyImpact.DEFAULT_AMP
                material: AudioMaterial = PyImpact.DEFAULT_MATERIAL
                resonance: float = PyImpact.DEFAULT_RESONANCE
                size: int = PyImpact.DEFAULT_SIZE
            # Get averages or maximums of each value.
            else:
                amp: float = round(sum(amps) / len(amps), 3)
                material: AudioMaterial = max(set(materials), key=materials.count)
                resonance: float = round(sum(resonances) / len(resonances), 3)
                size: int = int(sum(sizes) / len(sizes))
            derived_data[object_id] = ObjectAudioStatic(name=names[object_id],
                                                        mass=object_masses[object_id],
                                                        material=material,
                                                        bounciness=object_bouncinesses[object_id],
                                                        resonance=resonance,
                                                        size=size,
                                                        amp=amp,
                                                        object_id=object_id)
        # Add the derived data.
        for object_id in derived_data:
            self._static_audio_data[object_id] = derived_data[object_id]
        # Add robot joints.
        for joint_id in robot_joints:
            self._static_audio_data[joint_id] = ObjectAudioStatic(name=robot_joints[joint_id]["name"],
                                                                  mass=robot_joints[joint_id]["mass"],
                                                                  material=PyImpact.ROBOT_JOINT_MATERIAL,
                                                                  bounciness=PyImpact.ROBOT_JOINT_BOUNCINESS,
                                                                  resonance=PyImpact.DEFAULT_RESONANCE,
                                                                  size=PyImpact.DEFAULT_SIZE,
                                                                  amp=PyImpact.DEFAULT_AMP,
                                                                  object_id=joint_id)

    @staticmethod
    def _normalize_16bit_int(arr: np.array) -> np.array:
        """
        Convert numpy float array to normalized 16-bit integers.

        :param arr: Numpy float data to convert.

        :return: The converted numpy array.
        """

        normalized_floats = PyImpact._normalize_floats(arr)

        return (normalized_floats * 32767).astype(np.int16)

    @staticmethod
    def _normalize_floats(arr: np.array) -> np.array:
        """
        Normalize numpy array of float audio data.

        :param arr: Numpy float data to normalize.

        :return The normalized array.
        """

        if np.all(arr == 0):
            return arr
        else:
            return arr / np.abs(arr).max()

    def _end_scrape(self, scrape_key: Tuple[int, int]) -> None:
        """
        Clean up after a given scrape event has ended.

        :param: The scrape index key.
        """

        if scrape_key in self._scrape_events_count:
            del self._scrape_events_count[scrape_key]
        if scrape_key in self._scrape_summed_masters:
            del self._scrape_summed_masters[scrape_key]
        if scrape_key in self._scrape_start_velocities:
            del self._scrape_start_velocities[scrape_key]
        if scrape_key in self._scrape_previous_indices:
            del self._scrape_previous_indices[scrape_key]

    def _get_audio_command(self, audio_source_id: int, contact_points: np.array, sound: Base64Sound) -> dict:
        """
        :param audio_source_id: The audio source ID.
        :param contact_points: The collision contact points.
        :param sound: The Base64Sound object.

        :return: A command to play audio data.
        """

        point = np.mean(contact_points, axis=0)
        return {"$type": "play_audio_data" if not self.resonance_audio else "play_point_source_data",
                "id": audio_source_id,
                "position": {"x": float(point[0]), "y": float(point[1]), "z": float(point[2])},
                "num_frames": sound.length,
                "num_channels": CHANNELS,
                "frame_rate": SAMPLE_RATE,
                "wav_data": sound.wav_str}



import numpy as np
from tdw.output_data import Collision
from tdw.collision_data.collision_base import CollisionBase


class CollisionObjObj(CollisionBase):
    """
    A collision between two objects.
    """

    def __init__(self, collision: Collision):
        """
        :param collision: The collision output data.
        """

        super().__init__(collision=collision)
        """:field
        The relative velocity of the objects.
        """
        self.relative_velocity: np.array = np.array(collision.get_relative_velocity())



from typing import List
import re
from json import loads
from pathlib import Path
from pkg_resources import resource_filename
from tdw.add_ons.add_on import AddOn
from tdw.controller import Controller


class Floorplan(AddOn):
    """
    Initialize a scene populated by objects in pre-scripted layouts.

    There are four scenes (1, 2, 4, 5), each with three visual variants (a, b, c). Each scene has three different object layouts (0, 1, 2).

    ```python
    from tdw.controller import Controller
    from tdw.add_ons.floorplan import Floorplan

    c = Controller()
    f = Floorplan()
    f.init_scene(scene="1a", layout=0)
    c.add_ons.append(f)
    c.communicate([])
    ```
    """

    def __init__(self):
        """
        (no parameters)
        """

        super().__init__()
        self.initialized = True

    def init_scene(self, scene: str, layout: int) -> None:
        """
        Set commands to initialize the scene. On the next frame, the scene will be initialized.

        Valid scenes and layouts:

        | `scene` | `layout` |
        | --- | --- |
        | 1a, 1b, or 1c | 0, 1, or 2 |
        | 2a, 2b, or 2c | 0, 1, or 2 |
        | 4a, 4b, or 4c | 0, 1, or 2 |
        | 5a, 5b, or 5c | 0, 1, or 2 |

        Where `1a` is the scene number (1) and visual variant (a).

        :param scene: The name of the floorplan scene to load. Can be just the suffix of the scene to load (such as `"1a"`) or the full name of the scene (such as `"floorplan_1a"`).
        :param layout: The layout index.
        """

        # Example: "1a"
        if re.match(r"^([1245][abc])", scene) is not None:
            scene_name = f"floorplan_{scene}"
            scene_index = scene[0]
        elif re.match(r"^floorplan_([1245][abc])", scene) is not None:
            scene_name = scene
            scene_index = scene[-2]
        else:
            raise Exception(f"Invalid scene: {scene}")
        layout_index = str(layout)
        floorplans = loads(Path(resource_filename(__name__, "floorplan_layouts.json")).
                           read_text(encoding="utf-8"))
        if layout_index not in floorplans[scene_index]:
            raise Exception(f"Layout not found: {layout_index}")

        objects = floorplans[scene_index][layout_index]
        # Load the scene and add post-processing commands.
        self.commands = [Controller.get_add_scene(scene_name=scene_name),
                         {"$type": "set_aperture",
                          "aperture": 8.0},
                         {"$type": "set_focus_distance",
                          "focus_distance": 2.25},
                         {"$type": "set_post_exposure",
                          "post_exposure": 0.4},
                         {"$type": "set_ambient_occlusion_intensity",
                          "intensity": 0.175},
                         {"$type": "set_ambient_occlusion_thickness_modifier",
                          "thickness": 3.5}]
        # Add objects.
        for o in objects:
            object_id = Controller.get_unique_id()
            self.commands.extend(Controller.get_add_physics_object(model_name=o["name"],
                                                                   library=o["library"],
                                                                   position=o["position"],
                                                                   rotation=o["rotation"],
                                                                   scale_factor=o["scale_factor"],
                                                                   kinematic=o["kinematic"],
                                                                   gravity=o["gravity"],
                                                                   object_id=object_id))

    def get_initialization_commands(self) -> List[dict]:
        return []

    def on_send(self, resp: List[bytes]) -> None:
        pass



from tdw.output_data import EnvironmentCollision
from tdw.collision_data.collision_base import CollisionBase


class CollisionObjEnv(CollisionBase):
    """
    A collision between an object and the environment.
    """

    def __init__(self, collision: EnvironmentCollision):
        """
        :param collision: The collision output data.
        """

        super().__init__(collision=collision)
        """:field
        True if this is a collision with the floor.
        """
        self.floor: bool = collision.get_floor()



from json import loads
from requests import get
from typing import List
from pkg_resources import get_distribution
from packaging import version
from tdw.version import __version__
from tdw.release.build import Build


class PyPi:
    """
    Compare the version of the installed tdw Python module to the PyPi version.
    """

    @staticmethod
    def strip_post_release(v: str) -> str:
        """
        If the version number has a post-release suffix (a fourth number), strip it.

        :param v: The version number.

        :return: The version, stripped of the post-release suffix.
        """

        if len(v.split(".")) > 3:
            return '.'.join(v.split('.')[:3])
        else:
            return v

    @staticmethod
    def get_major_release(v: str) -> str:
        """
        :param v: The version number.

        :return: The major release number (example: in 1.7.0, the major release is 7).
        """

        return v.split(".")[1].strip()

    @staticmethod
    def _get_pypi_releases() -> List[str]:
        """
        :return: A list of all available PyPi releases.
        """

        resp = get("https://pypi.org/pypi/tdw/json")
        data = loads(resp.content)
        versions = list(data["releases"].keys())
        versions.sort(key=lambda s: list(map(int, s.split('.'))))
        return versions

    @staticmethod
    def get_pypi_version(truncate: bool = False) -> str:
        """
        :param truncate: If true, remove the post-release number (the fourth number) if there is one.

        :return: The newest available tdw release on PyPi.
        """

        # From the list of available versions, get the last one (the most recent).
        v = PyPi._get_pypi_releases()[-1]

        # Strip the post-release suffix.
        if truncate:
            return PyPi.strip_post_release(v)
        else:
            return v

    @staticmethod
    def get_installed_tdw_version(truncate: bool = False) -> str:
        """
        :param truncate: If true, remove the post-release number (the fourth number) if there is one.

        :return: The version of the tdw Python module installed on this machine.
        """

        v = get_distribution("tdw").version

        # Strip the post-release suffix.
        if truncate:
            return PyPi.strip_post_release(v)
        else:
            return v

    @staticmethod
    def get_latest_post_release(v: str) -> str:
        """
        :param v: A three-part version string, e.g. 1.6.1

        :return: The most up-to-date version or post-release of the tdw module on PyPi with `v`, e.g. 1.6.1.10
        """

        releases = PyPi._get_pypi_releases()
        releases = sorted([r for r in releases if r.startswith(v)],
                          key=lambda r: bytes([int(n) for n in r.split(".")]))
        if len(releases) == 0:
            return ""
        return releases[-1]

    @staticmethod
    def get_latest_minor_release(v: str) -> str:
        """
        :param v: The version number.

        :return: The most up-to-date version in this major release. (Example: if v == 1.5.0, this returns 1.5.5)
        """

        v = PyPi.strip_post_release(v)
        releases = PyPi._get_pypi_releases()
        # Sort the list by the byte array representation to put double-digit version numbers in the correct order.
        releases = sorted([r for r in releases if r.startswith("1." + PyPi.get_major_release(v))],
                          key=lambda r: bytes([int(n) for n in r.split(".")]))
        if len(releases) == 0:
            return ""
        return releases[-1]

    @staticmethod
    def required_tdw_version_is_installed(required_version: str, build_version: str, comparison: str = "==") -> bool:
        """
        Check whether the correct version of TDW is installed.
        This is useful for other modules such as the Magnebot API that rely on certain versions of TDW.

        :param required_version: The required version of TDW.
        :param build_version: The version of the build.
        :param comparison: The type of comparison. Options: "==", ">", ">=".

        :return: True if the installed tdw module is the correct version.
        """

        valid_comparisons: List[str] = ["==", ">", ">="]
        if comparison not in valid_comparisons:
            raise Exception(f"Invalid comparison {comparison}. Options are: {valid_comparisons}")

        ok: bool = True
        required_version = PyPi.strip_post_release(required_version)
        required_version_parsed = version.parse(required_version)
        installed_version_parsed = version.parse(__version__)
        if (comparison == "==" and required_version_parsed != installed_version_parsed) or \
                (comparison == ">" and installed_version_parsed <= required_version_parsed) or \
                (comparison == ">=" and installed_version_parsed < required_version_parsed):
            print(f"WARNING! You have tdw {__version__} but you need tdw {required_version}. "
                  f"To install the correct version:"
                  f"\n\tIf you installed tdw from the GitHub repo (pip3 install -e .): "
                  f"git checkout v{PyPi.strip_post_release(required_version)}"
                  f"\n\tIf you installed tdw from PyPi (pip3 install tdw): "
                  f"pip3 install tdw=={required_version}")
            ok = False
        build_version_parsed = version.parse(build_version)
        if (comparison == "==" and build_version_parsed != required_version_parsed) or \
                (comparison == ">" and build_version_parsed <= required_version_parsed) or \
                (comparison == ">=" and build_version_parsed < required_version_parsed):
            url, url_exists = Build.get_url(required_version, check_head=False)
            print(f"WARNING! You are using TDW build {build_version} but you need TDW build {required_version}. "
                  f"\n\tDownload and extract: {url}")
            ok = False
        return ok



from tdw.physics_audio.modes import Modes


class CollisionAudioInfo:
    """
    Class containing information about collisions required by PyImpact to determine the volume of impact sounds.
    """

    def __init__(self, obj1_modes: Modes, obj2_modes: Modes, amp: float = 0.5, init_speed: float = 1):
        """
        :param amp: Amplitude of the first collision (must be between 0 and 1).
        :param init_speed: The speed of the initial collision (all collisions will be scaled relative to this).
        :param obj1_modes: The object's modes.
        :param obj2_modes: The other object's modes.
        """

        """:field
        The collision counter.
        """
        self.count: int = 0
        """:field
        Amplitude of the first collision (must be between 0 and 1).
        """
        self.amp: float = amp
        """:field
        The speed of the initial collision.
        """
        self.init_speed: float = init_speed
        """:field
        The object's modes.
        """
        self.obj1_modes: Modes = obj1_modes
        """:field
        The other object's modes.
        """
        self.obj2_modes: Modes = obj2_modes

    def count_collisions(self) -> None:
        """
        Update the counter for how many times two objects have collided.
        """

        self.count += 1



import numpy as np
from tdw.output_data import StaticRobot


class NonMoving:
    """
    Static data for a non-moving object attached to a robot (i.e. a sub-object mesh of a limb).
    """

    def __init__(self, static_robot: StaticRobot, index: int):
        """
        :param static_robot: Static robot output data from the build.
        :param index: The index of this object in the list of non-moving objects.
        """

        """:field
        The ID of this object.
        """
        self.object_id: int = static_robot.get_non_moving_id(index)
        """:field
        The name of this object.
        """
        self.name: str = static_robot.get_non_moving_name(index)
        """:field
        The segmentation color of this joint as an `[r, g, b]` numpy array.
        """
        self.segmentation_color: np.array = np.array(static_robot.get_non_moving_segmentation_color(index))



from typing import List, Dict, Optional, Tuple
import numpy as np
from tdw.output_data import OutputData, Raycast, Overlap
from tdw.add_ons.add_on import AddOn
from tdw.scene_data.scene_bounds import SceneBounds


class OccupancyMap(AddOn):
    """
    An occupancy map is a numpy array that divides a TDW into a grid. Each cell is free (no objects), non-free (has objects), or is outside of the environment.

    Each element in the occupancy map can be one of three values:

    | Value | Meaning                                                      |
    | ----- | ------------------------------------------------------------ |
    | -1    | The cell is out of bounds (there is no floor).               |
    | 1     | The cell is occupied by at least one object, occupied by an environment object (such as a wall), or otherwise not navigable (blocked by other objects). |
    | 0     | The cell is unoccupied.                                      |
    """

    # The height from which rays will be cast.
    _RAYCAST_Y: float = 100

    def __init__(self, cell_size: float = 0.5):
        """
        :param cell_size: The diameter of each cell in meters.
        """

        super().__init__()
        """:field
        A 2-dimensional numpy array of the occupancy map. Each row corresponds to a worldspace x value and each column corresponds to a worldspace z value (see `get_occupancy_position(idx, idy)` below).
        """
        self.occupancy_map: Optional[np.array] = None
        """:field
        The [bounds of the scene](../scene_data/scene_bounds.md).
        """
        self.scene_bounds: Optional[SceneBounds] = None
        # The diameter of each cell in meters.
        self._cell_size: float = cell_size
        # The expected dimensions of the occupancy map array.
        self._occupancy_map_size: Tuple[int, int] = (0, 0)
        # Ignore these objects when generating the occupancy map.
        self._ignore_objects: List[int] = list()

    def get_initialization_commands(self) -> List[dict]:
        return [{"$type": "send_scene_regions"}]

    def on_send(self, resp: List[bytes]) -> None:
        def __get_islands() -> List[List[Tuple[int, int]]]:
            """
            :return: A list of all islands, i.e. continuous zones of traversability on the occupancy map.
            """

            # Positions that have been reviewed so far.
            traversed: List[Tuple[int, int]] = []
            islands: List[List[Tuple[int, int]]] = list()

            for ox, oy in np.ndindex(self.occupancy_map.shape):
                op = (ox, oy)
                if op in traversed:
                    continue
                # Fill the island (a continuous zone) that position `p` belongs to.
                to_check: List[tuple] = [op]
                island: List[Tuple[int, int]] = list()
                while len(to_check) > 0:
                    # Check the next position.
                    op = to_check.pop(0)
                    if op[0] < 0 or op[0] >= self.occupancy_map.shape[0] or op[1] < 0 or \
                            op[1] >= self.occupancy_map.shape[1] or \
                            self.occupancy_map[op[0]][op[1]] != 0 or op in island:
                        continue
                    # Mark the position as traversed.
                    island.append(op)
                    # Check these neighbors.
                    px, py = op
                    to_check.extend([(px, py + 1),
                                     (px + 1, py + 1),
                                     (px + 1, py),
                                     (px + 1, py - 1),
                                     (px, py - 1),
                                     (px - 1, py - 1),
                                     (px - 1, py),
                                     (px - 1, py + 1)])
                if len(island) > 0:
                    for island_position in island:
                        traversed.append(island_position)
                    islands.append(island)
            return islands

        # Set the scene bounds.
        if self.scene_bounds is None:
            self.scene_bounds = SceneBounds(resp=resp)
        if self.occupancy_map is None:
            # Generate the occupancy map.
            self.occupancy_map = np.zeros(shape=(self._occupancy_map_size[0] + 1, self._occupancy_map_size[1] + 1),
                                          dtype=int)
            # Get all of the positions that are actually in the environment.
            hit_env: Dict[int, bool] = dict()
            # Get all of the overlaps to determine if there was an object.
            hit_obj: Dict[int, bool] = dict()
            # The IDs of each object in the overlap.
            hit_obj_ids: Dict[int, np.array] = dict()
            hit_walls: Dict[int, bool] = dict()
            for i in range(len(resp) - 1):
                r_id = OutputData.get_data_type_id(resp[i])
                if r_id == "rayc":
                    raycast = Raycast(resp[i])
                    hit_env[raycast.get_raycast_id()] = raycast.get_hit()
                elif r_id == "over":
                    overlap = Overlap(resp[i])
                    overlap_id = overlap.get_id()
                    hit_obj[overlap_id] = len(overlap.get_object_ids()) > 0
                    hit_obj_ids[overlap_id] = overlap.get_object_ids()
                    hit_walls[overlap_id] = overlap.get_walls()

            for cast_id in hit_env:
                idx = cast_id % 10000
                idz = int((cast_id - (cast_id % 10000)) / 10000)
                # The position is outside of the environment.
                if not hit_env[cast_id]:
                    self.occupancy_map[idx][idz] = -1
                # The position is occupied by at least one object that we aren't ignoring.
                elif hit_walls[cast_id] or (hit_obj[cast_id] and len([o for o in hit_obj_ids[cast_id] if o not in self._ignore_objects]) > 0):
                    self.occupancy_map[idx][idz] = 1
                # The position is free.
                else:
                    self.occupancy_map[idx][idz] = 0
            # Assume that the edges of the occupancy map are out of bounds.
            for ix, iy in np.ndindex(self.occupancy_map.shape):
                if ix == 0 or ix == self.occupancy_map.shape[0] - 1 or iy == 0 or \
                        iy == self.occupancy_map.shape[1] - 1:
                    self.occupancy_map[ix][iy] = -1
            # Sort the free positions of the occupancy map into continuous "islands".
            # Then, sort that list of lists by length.
            # The longest list is the biggest "island" i.e. the navigable area.
            non_navigable = list(sorted(__get_islands(), key=len))[:-1]
            # Record non-navigable positions.
            for n in non_navigable:
                for p in n:
                    self.occupancy_map[p[0]][p[1]] = -1

    def generate(self, ignore_objects: List[int] = None) -> None:
        """
        Generate an occupancy map.
        This function should only be called at least one controller.communicate() call after adding this add-on.
        The OccupancyMap then requires one more controller.communicate() call to create the occupancy map.
        (See the example at the top of this document.)

        :param ignore_objects: If not None, ignore these objects when determining if a cell is free or non-free.
        """

        if not self.initialized:
            raise Exception("Can't generate an occupancy map because this add-on hasn't initialized.\n"
                            "Wait at least one controller.communicate() call before calling occupancy_map.generate()")
        self.occupancy_map = None
        if ignore_objects is None:
            self._ignore_objects.clear()
        else:
            self._ignore_objects = ignore_objects
        # Spherecast to each point.
        x = self.scene_bounds.x_min
        idx = 0
        idz = 0
        capsule_half_height = (self.scene_bounds.y_max - self.scene_bounds.y_min) / 2
        while x < self.scene_bounds.x_max:
            z = self.scene_bounds.z_min
            idz = 0
            while z < self.scene_bounds.z_max:
                # Create an overlap sphere to determine if the cell is occupied.
                # Cast a ray to determine if the cell has a floor.
                cast_id = idx + (idz * 10000)
                self.commands.extend([{"$type": "send_overlap_capsule",
                                       "end": {"x": x, "y": capsule_half_height, "z": z},
                                       "radius": self._cell_size / 2,
                                       "position": {"x": x, "y": -capsule_half_height, "z": z},
                                       "id": cast_id},
                                      {"$type": "send_raycast",
                                       "origin": {"x": x, "y": OccupancyMap._RAYCAST_Y, "z": z},
                                       "destination": {"x": x, "y": -1, "z": z},
                                       "id": cast_id}])
                z += self._cell_size
                idz += 1
            x += self._cell_size
            idx += 1
        self._occupancy_map_size = (idx, idz)

    def get_occupancy_position(self, i: int, j: int) -> Tuple[float, float]:
        """
        Convert occupancy map indices to worldspace coordinates.
        This function can only be sent after first calling `self.generate()` and waiting at least one `controller.communicate()` call.:

        ```python
        from tdw.controller import Controller
        from tdw.tdw_utils import TDWUtils
        from tdw.add_ons.occupancy_map import OccupancyMap

        c = Controller(launch_build=False)
        o = OccupancyMap(cell_size=0.5)
        c.add_ons.append(o)
        c.communicate(TDWUtils.create_empty_room(12, 12))
        o.generate()
        c.communicate([])
        print(o.get_occupancy_position(4, 5))  # (-3.5, -3.0)
        c.communicate({"$type": "terminate"})
        ```

        :param i: The column index of self.occupancy_map
        :param j: The row index of self.occupancy_map.

        :return: A tuple: `self.occupancy_map[i][j]` converted into `(x, z)` worldspace coordinates.
        """

        if self.scene_bounds is None:
            raise Exception("The scene bounds haven't been generated and initialized (see documentation).")

        return self.scene_bounds.x_min + (i * self._cell_size), self.scene_bounds.z_min + (j * self._cell_size)

    def show(self) -> None:
        """
        Visualize the occupancy map by adding blue squares into the scene to mark free spaces.
        These blue squares don't interact with the physics engine.
        """

        if self.occupancy_map is None:
            raise Exception("The occupancy map hasn't been generated and initialized (see documentation).")
        for idx, idy in np.ndindex(self.occupancy_map.shape):
            if self.occupancy_map[idx][idy] != 0:
                continue
            x, z = self.get_occupancy_position(idx, idy)
            self.commands.append({"$type": "add_position_marker",
                                  "position": {"x": x, "y": 0.05, "z": z},
                                  "scale": self._cell_size * 0.9,
                                  "color": {"r": 0, "g": 0, "b": 1, "a": 1},
                                  "shape": "square"})

    def hide(self) -> None:
        self.commands.append({"$type": "remove_position_markers"})



from typing import Dict
from tdw.physics_audio.audio_material import AudioMaterial


# Density per audio material.
DENSITIES: Dict[AudioMaterial, float] = {AudioMaterial.ceramic: 2180,
                                         AudioMaterial.glass: 2500,
                                         AudioMaterial.stone: 2000,
                                         AudioMaterial.metal: 8450,
                                         AudioMaterial.wood_hard: 1200,
                                         AudioMaterial.wood_medium: 700,
                                         AudioMaterial.wood_soft: 400,
                                         AudioMaterial.fabric: 1540,
                                         AudioMaterial.leather: 860,
                                         AudioMaterial.plastic_hard: 1150,
                                         AudioMaterial.plastic_soft_foam: 285,
                                         AudioMaterial.rubber: 1522,
                                         AudioMaterial.paper: 1200,
                                         AudioMaterial.cardboard: 698}
# Dynamic friction per audio material.
DYNAMIC_FRICTION: Dict[AudioMaterial, float] = {AudioMaterial.ceramic: 0.47,
                                                AudioMaterial.wood_hard: 0.35,
                                                AudioMaterial.wood_medium: 0.35,
                                                AudioMaterial.wood_soft: 0.35,
                                                AudioMaterial.cardboard: 0.45,
                                                AudioMaterial.paper: 0.47,
                                                AudioMaterial.glass: 0.65,
                                                AudioMaterial.fabric: 0.65,
                                                AudioMaterial.leather: 0.4,
                                                AudioMaterial.stone: 0.7,
                                                AudioMaterial.rubber: 0.75,
                                                AudioMaterial.plastic_hard: 0.3,
                                                AudioMaterial.plastic_soft_foam: 0.45,
                                                AudioMaterial.metal: 0.43}
# Static friction per audio material.
STATIC_FRICTION: Dict[AudioMaterial, float] = {AudioMaterial.ceramic: 0.47,
                                               AudioMaterial.wood_hard: 0.37,
                                               AudioMaterial.wood_medium: 0.37,
                                               AudioMaterial.wood_soft: 0.37,
                                               AudioMaterial.cardboard: 0.48,
                                               AudioMaterial.paper: 0.5,
                                               AudioMaterial.glass: 0.68,
                                               AudioMaterial.fabric: 0.67,
                                               AudioMaterial.leather: 0.43,
                                               AudioMaterial.stone: 0.72,
                                               AudioMaterial.rubber: 0.8,
                                               AudioMaterial.plastic_hard: 0.35,
                                               AudioMaterial.plastic_soft_foam: 0.47,
                                               AudioMaterial.metal: 0.47}



from abc import ABC, abstractmethod
from typing import List, Dict, Optional
from overrides import final
import numpy as np
from tdw.robot_data.robot_static import RobotStatic
from tdw.robot_data.robot_dynamic import RobotDynamic
from tdw.add_ons.add_on import AddOn


class RobotBase(AddOn, ABC):
    """
    Abstract base class for robots.
    """

    """:class_var
    If a joint has moved less than this many degrees (revolute or spherical) or meters (prismatic) since the previous frame, it is considered to be not moving for the purposes of determining which joints are moving.
    """
    NON_MOVING: float = 0.001

    def __init__(self, robot_id: int = 0, position: Dict[str, float] = None, rotation: Dict[str, float] = None):
        """
        :param robot_id: The ID of the robot.
        :param position: The position of the robot. If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        :param rotation: The rotation of the robot in Euler angles (degrees). If None, defaults to `{"x": 0, "y": 0, "z": 0}`.
        """

        super().__init__()

        if position is None:
            """:field
            The initial position of the robot.
            """
            self.initial_position: Dict[str, float] = {"x": 0, "y": 0, "z": 0}
        else:
            self.initial_position: Dict[str, float] = position
        if rotation is None:
            """:field
            The initial rotation of the robot.
            """
            self.initial_rotation: Dict[str, float] = {"x": 0, "y": 0, "z": 0}
        else:
            self.initial_rotation: Dict[str, float] = rotation
        """:field
        The ID of this robot.
        """
        self.robot_id: int = robot_id
        """:field
        Static robot data.
        """
        self.static: Optional[RobotStatic] = None
        """:field
        Dynamic robot data.
        """
        self.dynamic: Optional[RobotDynamic] = None

    def get_initialization_commands(self) -> List[dict]:
        """
        This function gets called exactly once per add-on. To re-initialize, set `self.initialized = False`.

        :return: A list of commands that will initialize this add-on.
        """

        self.static = None
        return [self._get_add_robot_command(),
                {"$type": "send_static_robots",
                 "frequency": "once"},
                {"$type": "send_robots",
                 "frequency": "always"}]

    def joints_are_moving(self, joint_ids: List[int] = None) -> bool:
        """
        :param joint_ids: A list of joint IDs to check for movement. If `None`, check all joints for movement.

        :return: True if the joints are moving.
        """

        if joint_ids is None:
            joint_ids = self.dynamic.joints.keys()
        for joint_id in joint_ids:
            if self.dynamic.joints[joint_id].moving:
                return True
        return False

    def reset(self, position: Dict[str, float] = None, rotation: Dict[str, float] = None) -> None:
        """
        Reset the robot.

        :param position: The position of the robot.
        :param rotation: The rotation of the robot.
        """

        self.initialized = False
        self.static = None
        self.dynamic = None
        self.commands.clear()
        if position is not None:
            self.initial_position = position
        if rotation is not None:
            self.initial_rotation = rotation

    def on_send(self, resp: List[bytes]) -> None:
        """
        This is called after commands are sent to the build and a response is received.

        Use this function to send commands to the build on the next frame, given the `resp` response.
        Any commands in the `self.commands` list will be sent on the next frame.

        :param resp: The response from the build.
        """

        if self.static is None:
            self._cache_static_data(resp=resp)
        self._set_dynamic_data(resp=resp)

    @final
    def before_send(self, commands: List[dict]) -> None:
        """
        This is called before sending commands to the build. By default, this function doesn't do anything.

        :param commands: The commands that are about to be sent to the build.
        """

        pass

    @abstractmethod
    def _cache_static_data(self, resp: List[bytes]) -> None:
        """
        Cache static output data.

        :param resp: The response from the build.
        """

        raise Exception()

    @abstractmethod
    def _set_dynamic_data(self, resp: List[bytes]) -> None:
        """
        Set dynamic data per frame.

        :param resp: The response from the build.
        """

        raise Exception()

    @abstractmethod
    def _get_add_robot_command(self) -> dict:
        """
        :return: A command to add the robot.
        """

        raise Exception()

    @final
    def _set_joints_moving(self, dynamic: RobotDynamic) -> RobotDynamic:
        """
        Set the state of the dynamic data regarding whether the joints are currently moving.

        :param dynamic: The dynamic data.

        :return: The updated dynamic data.
        """

        if self.dynamic is not None:
            # Check which joints are still moving.
            for joint_id in dynamic.joints:
                dynamic.joints[joint_id].moving = False
                for angle_0, angle_1 in zip(self.dynamic.joints[joint_id].angles, dynamic.joints[joint_id].angles):
                    if np.linalg.norm(angle_1 - angle_0) > RobotBase.NON_MOVING:
                        dynamic.joints[joint_id].moving = True
                        break
        else:
            for joint_id in dynamic.joints:
                dynamic.joints[joint_id].moving = True
        return dynamic



from abc import ABC, abstractmethod
from typing import List
from tdw.librarian import ModelRecord


class ModelTest(ABC):
    """
    Run a test on a model.
    """

    def __init__(self, record: ModelRecord):
        """
        :param record: The model record.
        """

        # The model record.
        self._record: ModelRecord = record
        """:field
        A list of report strings after running the test.
        """
        self.reports: List[str] = list()
        """:field
        If True, the test is done.
        """
        self.done: bool = False

    @abstractmethod
    def start(self) -> List[dict]:
        """
        :return: A list of commands to start the test.
        """

        raise Exception()

    @abstractmethod
    def on_send(self, resp: List[bytes]) -> List[dict]:
        """
        :param resp: The response from the build.

        :return: A list of commands to continue or end the test.
        """

        raise Exception()



from typing import Dict, List
import numpy as np
from tdw.output_data import OutputData, Transforms, Rigidbodies, Bounds, SegmentationColors, Categories, \
    StaticRigidbodies
from tdw.add_ons.add_on import AddOn
from tdw.object_data.object_static import ObjectStatic
from tdw.object_data.transform import Transform
from tdw.object_data.rigidbody import Rigidbody
from tdw.object_data.bound import Bound


class _StaticRigidbody:
    """
    Temporary structure for static rigidbody data.
    """

    def __init__(self, mass: float, kinematic: bool, dynamic_friction, static_friction: float, bounciness: float):
        """
        :param mass: The mass of the object.
        :param dynamic_friction: The dynamic friction of the object.
        :param static_friction: The static friction of the object.
        :param bounciness: The bounciness of the object.
        :param kinematic: If True, this object is kinematic, and won't respond to physics.
        """

        """:field
        If True, this object is kinematic, and won't respond to physics. 
        Examples: a painting hung on a wall or built-in furniture like a countertop.
        """
        self.kinematic = kinematic
        """:field
        The mass of the object.
        """
        self.mass = mass
        """:field
        The dynamic friction of the object.
        """
        self.dynamic_friction: float = dynamic_friction
        """:field
        The static friction of the object.
        """
        self.static_friction: float = static_friction
        """:field
        The bounciness of the object.
        """
        self.bounciness: float = bounciness


class ObjectManager(AddOn):
    """
    A simple manager class for objects in the scene. This add-on can cache static object data (name, ID, etc.) and record dynamic data (position, velocity, etc.) per frame.

    ## Usages constraints:

    - This add-on assumes that this is a PhysX simulation, as opposed to a simulation with physics disabled or a Flex simulation.
    - This add-on will record data for *all* objects in the scene. If you only need data for specific objects, you should use low-level TDW commands.
    - By default, this add-on will record [transform data](../object_data/transform.md) but not [rigidbody data](../object_data/rigidbody.md) or [bounds data](../object_data/bound.md). You can set which data the add-on will record in the constructor, but be aware that this can slow down the simulation.

    ## Example usage

    ```python
    import numpy as np
    from tdw.controller import Controller
    from tdw.tdw_utils import TDWUtils
    from tdw.librarian import ModelLibrarian
    from tdw.add_ons.object_manager import ObjectManager


    c = Controller()
    c.model_librarian = ModelLibrarian("models_special.json")
    # Create the object manager.
    om = ObjectManager()
    c.add_ons.append(om)
    commands = [TDWUtils.create_empty_room(100, 100)]
    # The starting height of the objects.
    y = 10
    # The radius of the circle of objects.
    r = 7.0
    # Get all points within the circle defined by the radius.
    p0 = np.array((0, 0))
    o_id = 0
    for x in np.arange(-r, r, 1):
        for z in np.arange(-r, r, 1):
            p1 = np.array((x, z))
            dist = np.linalg.norm(p0 - p1)
            if dist < r:
                commands.extend([c.get_add_object("prim_cone",
                                                  object_id=o_id,
                                                  position={"x": x, "y": y, "z": z},
                                                  rotation={"x": 0, "y": 0, "z": 180})])
                o_id += 1
    pass
    c.communicate(commands)
    for i in range(1000):
        for object_id in om.transforms:
            print(object_id, om.transforms[object_id].position)
        c.communicate([])
    c.communicate({"$type": "terminate"})
    ```
    """
    def __init__(self, transforms: bool = True, rigidbodies: bool = False, bounds: bool = False):
        """
        :param transforms: If True, record the [transform data](../object_data/transform.md) of each object in the scene.
        :param rigidbodies: If True, record the [rigidbody data](../object_data/rigidbody.md) of each rigidbody object in the scene.
        :param bounds: If True, record the [bounds data](../object_data/bound.md) of each object in the scene.
        """

        super().__init__()
        self._cached_static_data: bool = False
        self._send_transforms: str = "always" if transforms else "never"
        self._send_rigidbodies: str = "always" if rigidbodies else "never"
        self._send_bounds: str = "always" if bounds else "once"
        """:field
        [The static object data.](../object_data/object_static.md) Key = The ID of the object.
        """
        self.objects_static: Dict[int, ObjectStatic] = dict()
        """:field
        The segmentation color per category as use in the _category image pass. Key = The category. Value = The color as an `[r, g, b]` numpy array.
        """
        self.categories: Dict[str, np.array] = dict()
        """:field
        The [transform data](../object_data/transform.md) for each object on the scene on this frame. Key = The object ID. If `transforms=False` in the constructor, this dictionary will be empty.
        """
        self.transforms: Dict[int, Transform] = dict()
        """:field
        The [rigidbody data](../object_data/rigidbody.md) for each rigidbody object on the scene on this frame. Key = The object ID. If `rigidbodies=False` in the constructor, this dictionary will be empty.
        """
        self.rigidbodies: Dict[int, Rigidbody] = dict()
        """:field
        The [bounds data](../object_data/bound.md) for each object on the scene on this frame. Key = The object ID. If `bounds=False` in the constructor, this dictionary will be empty.
        """
        self.bounds: Dict[int, Bound] = dict()

    def get_initialization_commands(self) -> List[dict]:
        return [{"$type": "send_segmentation_colors"},
                {"$type": "send_categories"},
                {"$type": "send_static_rigidbodies"},
                {"$type": "send_rigidbodies",
                 "frequency": self._send_rigidbodies},
                {"$type": "send_bounds",
                 "frequency": self._send_bounds},
                {"$type": "send_transforms",
                 "frequency": self._send_transforms}]

    def on_send(self, resp: List[bytes]) -> None:
        # Cache static data.
        if not self._cached_static_data:
            self._cached_static_data = True
            # Sort the static output data by object ID.
            segmentation_colors: Dict[int, np.array] = dict()
            names: Dict[int, str] = dict()
            static_rigidbodies: Dict[int, _StaticRigidbody] = dict()
            sizes: Dict[int, np.array] = dict()
            categories: Dict[int, str] = dict()
            for i in range(len(resp) - 1):
                r_id = OutputData.get_data_type_id(resp[i])
                # Get the name and the segmentation color.
                if r_id == "segm":
                    segm = SegmentationColors(resp[i])
                    for j in range(segm.get_num()):
                        object_id = segm.get_object_id(j)
                        segmentation_colors[object_id] = np.array(segm.get_object_color(j))
                        names[object_id] = segm.get_object_name(j).lower()
                        categories[object_id] = segm.get_object_category(j)
                elif r_id == "boun":
                    boun = Bounds(resp[i])
                    for j in range(boun.get_num()):
                        sizes[boun.get_id(j)] = np.array([float(np.abs(boun.get_right(j)[0] - boun.get_left(j)[0])),
                                                          float(np.abs(boun.get_top(j)[1] - boun.get_bottom(j)[1])),
                                                          float(np.abs(boun.get_front(j)[2] - boun.get_back(j)[2]))])
                elif r_id == "srig":
                    srig = StaticRigidbodies(resp[i])
                    for j in range(srig.get_num()):
                        static_rigidbodies[srig.get_id(j)] = _StaticRigidbody(mass=srig.get_mass(j),
                                                                              kinematic=srig.get_kinematic(j),
                                                                              dynamic_friction=srig.get_dynamic_friction(j),
                                                                              static_friction=srig.get_static_friction(j),
                                                                              bounciness=srig.get_bounciness(j))
                elif r_id == "cate":
                    cate = Categories(resp[i])
                    for j in range(cate.get_num_categories()):
                        self.categories[cate.get_category_name(j)] = np.array(cate.get_category_color(j))
            # Cache the sorted data.
            for object_id in segmentation_colors:
                self.objects_static[object_id] = ObjectStatic(object_id=object_id,
                                                              name=names[object_id],
                                                              segmentation_color=segmentation_colors[object_id],
                                                              mass=static_rigidbodies[object_id].mass,
                                                              kinematic=static_rigidbodies[object_id].kinematic,
                                                              dynamic_friction=static_rigidbodies[object_id].dynamic_friction,
                                                              static_friction=static_rigidbodies[object_id].static_friction,
                                                              bounciness=static_rigidbodies[object_id].bounciness,
                                                              size=sizes[object_id],
                                                              category=categories[object_id])
        # Set dynamic data.
        self.transforms.clear()
        self.rigidbodies.clear()
        self.bounds.clear()
        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "tran":
                tran = Transforms(resp[i])
                for j in range(tran.get_num()):
                    self.transforms[tran.get_id(j)] = Transform(position=np.array(tran.get_position(j)),
                                                                rotation=np.array(tran.get_rotation(j)),
                                                                forward=np.array(tran.get_forward(j)))
            elif r_id == "rigi":
                rigi = Rigidbodies(resp[i])
                for j in range(rigi.get_num()):
                    self.rigidbodies[rigi.get_id(j)] = Rigidbody(velocity=rigi.get_velocity(j),
                                                                 angular_velocity=rigi.get_angular_velocity(j),
                                                                 sleeping=rigi.get_sleeping(j))
            elif r_id == "boun":
                boun = Bounds(resp[i])
                for j in range(boun.get_num()):
                    self.bounds[boun.get_id(j)] = Bound(front=np.array(boun.get_front(j)),
                                                        back=np.array(boun.get_back(j)),
                                                        left=np.array(boun.get_left(j)),
                                                        right=np.array(boun.get_right(j)),
                                                        top=np.array(boun.get_top(j)),
                                                        bottom=np.array(boun.get_bottom(j)),
                                                        center=np.array(boun.get_center(j)))

    def reset(self) -> None:
        """
        Reset the cached static data. Call this when resetting the scene.
        """

        self._cached_static_data = False
        self.objects_static.clear()
        self.categories.clear()
        self.initialized = False



from typing import List, Dict
from abc import ABC, abstractmethod
import numpy as np
from tdw.add_ons.model_verifier.model_tests.model_test import ModelTest
from tdw.librarian import ModelRecord
from tdw.tdw_utils import TDWUtils
from tdw.output_data import OutputData, Images


class RotateObjectTest(ModelTest, ABC):
    """
    These tests add an object and then rotate it.
    """

    """:class_var
    The ID of the object.
    """
    OBJECT_ID: int = 0
    """:class_var
    Rotate by this many degrees per frame.
    """
    DELTA_THETA: int = 15
    """:class_var
    The Unity pink color.
    """
    PINK: tuple = (255, 0, 255)
    """:class_var
    Look at this position.
    """
    LOOK_AT: Dict[str, float] = {"x": 0, "y": 0.5, "z": 0}
    """:class_var
    The position of the avatar.
    """
    AVATAR_POSITION: Dict[str, float] = {"x": 1.75, "y": 0.5, "z": 0}

    def __init__(self, record: ModelRecord):
        """
        :param record: The model record.
        """

        super().__init__(record=record)
        self._axis: str = "yaw"
        self._angle: int = 0

    def start(self) -> List[dict]:
        """
        :return: A list of commands to start the test.
        """

        scale = TDWUtils.get_unit_scale(self._record)
        # Create the scene. Add the avatar. Add the object.
        return [{"$type": "send_images",
                "frequency": "always"},
                {"$type": "add_object",
                 "name": self._record.name,
                 "url": self._record.get_url(),
                 "position": {"x": 0, "y": 0, "z": 0},
                 "scale_factor": self._record.scale_factor,
                 "id": RotateObjectTest.OBJECT_ID},
                {"$type": "scale_object",
                 "id": RotateObjectTest.OBJECT_ID,
                 "scale_factor": {"x": scale, "y": scale, "z": scale}}]

    def on_send(self, resp: List[bytes]) -> List[dict]:
        """
        :param resp: The response from the build.

        :return: A list of commands to continue or end the test.
        """

        for i in range(len(resp) - 1):
            r_id = OutputData.get_data_type_id(resp[i])
            if r_id == "imag":
                self._read_images(Images(resp[i]))
                break
        # Reading the images can cause the test to finish early.
        if self.done:
            return []
        # Either end the test or reset the angle and start rotating around a new axis.
        elif self._angle >= 360:
            if self._axis == "yaw":
                self._axis = "roll"
                self._angle = 0
                return [{"$type": "teleport_avatar_to",
                         "position": RotateObjectTest.AVATAR_POSITION},
                        {"$type": "look_at_position",
                         "position": RotateObjectTest.LOOK_AT},]
            else:
                self.done = True
                return []
        # Continue to rotate.
        else:
            self._angle += RotateObjectTest.DELTA_THETA
            rad = np.radians(self._angle)
            if self._axis == "yaw":
                x = np.cos(rad) * RotateObjectTest.AVATAR_POSITION["x"] - np.sin(rad) * RotateObjectTest.AVATAR_POSITION["z"]
                y = RotateObjectTest.AVATAR_POSITION["y"]
                z = np.sin(rad) * RotateObjectTest.AVATAR_POSITION["x"] + np.cos(rad) * RotateObjectTest.AVATAR_POSITION["z"]
            else:
                x = np.cos(rad) * RotateObjectTest.AVATAR_POSITION["x"] - np.sin(rad) * RotateObjectTest.AVATAR_POSITION["z"]
                y = (np.sin(rad) * RotateObjectTest.AVATAR_POSITION["x"] + np.cos(rad) * RotateObjectTest.AVATAR_POSITION["z"]) + RotateObjectTest.AVATAR_POSITION["y"]
                z = RotateObjectTest.AVATAR_POSITION["z"]

            return [{"$type": "teleport_avatar_to",
                     "position": {"x": x, "y": y, "z": z}},
                    {"$type": "look_at_position",
                     "position": RotateObjectTest.LOOK_AT}]

    @abstractmethod
    def _read_images(self, images: Images) -> None:
        """
        Read image data.

        :param images: The image data.
        """

        raise Exception()

    @staticmethod
    def _get_end_commands() -> List[dict]:
        """
        :return: A list of commands to end to test.
        """

        return [{"$type": "destroy_object",
                 "id": RotateObjectTest.OBJECT_ID},
                {"$type": "unload_asset_bundles"}]



from abc import ABC
from typing import List, Union
import numpy as np
from tdw.output_data import Collision, EnvironmentCollision


class CollisionBase(ABC):
    """
    Abstract base class for collision data.
    """

    def __init__(self, collision: Union[Collision, EnvironmentCollision]):
        """
        :param collision: The collision output data.
        """

        """:field
        The contact point positions.
        """
        self.points: List[np.array] = list()
        """:field
        The contact point normals.
        """
        self.normals: List[np.array] = list()
        for i in range(collision.get_num_contacts()):
            self.points.append(np.array(collision.get_contact_point(i)))
            self.normals.append(np.array(collision.get_contact_normal(i)))
        """:field
        The state of the collision.
        """
        self.state: str = collision.get_state()



import math
import numpy as np


class Modes:
    """
    Resonant mode properties: Frequencies, powers, and times.
    """

    def __init__(self, frequencies: np.array, powers: np.array, decay_times: np.array):
        """
        :param frequencies: A numpy array of mode frequencies in Hz.
        :param powers: A numpy array of mode onset powers in dB re 1.
        :param decay_times: A numpy array of mode decay times i.e. the time in ms it takes for each mode to decay 60dB from its onset power.
        """

        """:field
        A numpy array of mode frequencies in Hz.
        """
        self.frequencies: np.array = frequencies
        """:field
        A numpy array of mode onset powers in dB re 1.
        """
        self.powers: np.array = powers
        """:field
        A numpy array of mode decay times i.e. the time in ms it takes for each mode to decay 60dB from its onset power.
        """
        self.decay_times: np.array = decay_times

    def sum_modes(self, fs: int = 44100, resonance: float = 1.0) -> np.array:
        """
        Create mode time-series from mode properties and sum them together.

        :param fs: The framerate.
        :param resonance: The object resonance.

        :return A synthesized sound.
        """

        synth_sound: np.array = np.empty
        # Scroll through modes.
        for i in range(len(self.frequencies)):
            H_dB = 80 + self.powers[i]
            L_ms = self.decay_times[i] * H_dB / 60
            mLen = math.ceil(L_ms / 1e3 * fs)
            # if mode length is greater than the current time-series we have had make our time series longer
            max_len = mLen
            if mLen > max_len:
                max_len = mLen
            tt = np.arange(0, max_len) / fs
            # synthesize a sinusoid
            mode = np.cos(2 * math.pi * self.frequencies[i] * tt)
            mode = mode * (10 ** (self.powers[i] / 20))
            dcy = tt * (60 / (self.decay_times[i] * resonance / 1e3))
            env = 10 ** (-dcy / 20)
            mode = mode * env
            if i == 0:
                synth_sound = mode
            else:
                synth_sound = Modes.mode_add(synth_sound, mode)
        return synth_sound

    @staticmethod
    def mode_add(a: np.array, b: np.array) -> np.array:
        """
        Add together numpy arrays of different lengths by zero-padding the shorter.

        :param a: The first array.
        :param b: The second array.

        :return The summed modes.
        """

        if len(a) < len(b):
            c = b.copy()
            c[:len(a)] += a
        else:
            c = a.copy()
            c[:len(b)] += b
        return c
